# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from pytest import raises, mark
from numpy import loadtxt, arange
from numpy.testing import assert_array_equal

from hyperspy.signals import Signal2D
from lumispy.signals import LumiSpectrum
from lumispy import savetxt


@mark.parametrize("axes", (True, False))
def test_to_array_spectrum(axes):
    s = LumiSpectrum(arange(5))
    a = s.to_array(axes=axes)
    if axes:
        assert_array_equal(s.axes_manager[0].axis, a[:, 0])
    else:
        assert_array_equal(s.data, a)


@mark.parametrize("axes", (True, False))
@mark.parametrize("transpose", (True, False))
def test_to_array_linescan(axes, transpose):
    s = LumiSpectrum(arange(20).reshape((4, 5)))
    a = s.to_array(axes=axes, transpose=transpose)
    if axes:
        if transpose:
            assert_array_equal(s.axes_manager[0].axis, a[0, 1:])
            assert_array_equal(s.axes_manager[1].axis, a[1:, 0])
            assert_array_equal(s.data.T, a[1:, 1:])
        else:
            assert_array_equal(s.axes_manager[1].axis, a[0, 1:])
            assert_array_equal(s.axes_manager[0].axis, a[1:, 0])
            assert_array_equal(s.data, a[1:, 1:])
    else:
        if transpose:
            assert_array_equal(s.data.T, a)
        else:
            assert_array_equal(s.data, a)


@mark.parametrize("axes", (True, False))
def test_savetxt_spectrum(axes, tmp_path):
    s = LumiSpectrum(arange(5))
    fname = tmp_path / "test.txt"
    s.savetxt(fname, axes=axes)
    s2 = loadtxt(fname)
    if axes:
        assert_array_equal(s.axes_manager[0].axis, s2[:, 0])
    else:
        assert_array_equal(s.data, s2)


@mark.parametrize("axes", (True, False))
def test_savetxt_navigate1D(axes, tmp_path):
    s = LumiSpectrum(arange(5))
    s.axes_manager[0].navigate = True
    fname = tmp_path / "test.txt"
    s.savetxt(fname, axes=axes)
    s2 = loadtxt(fname)
    if axes:
        assert_array_equal(s.axes_manager[0].axis, s2[:, 0])
    else:
        assert_array_equal(s.data, s2)


@mark.parametrize("axes", (True, False))
@mark.parametrize("transpose", (True, False))
def test_savetxt_linescan(axes, transpose, tmp_path):
    s = LumiSpectrum(arange(20).reshape((4, 5)))
    fname = tmp_path / "test.txt"
    s.savetxt(fname, axes=axes, transpose=transpose)
    s2 = loadtxt(fname)
    if axes:
        if transpose:
            assert_array_equal(s.axes_manager[0].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[1].axis, s2[1:, 0])
            assert_array_equal(s.data.T, s2[1:, 1:])
        else:
            assert_array_equal(s.axes_manager[1].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[0].axis, s2[1:, 0])
            assert_array_equal(s.data, s2[1:, 1:])
    else:
        if transpose:
            assert_array_equal(s.data.T, s2)
        else:
            assert_array_equal(s.data, s2)


@mark.parametrize("axes", (True, False))
@mark.parametrize("transpose", (True, False))
def test_savetxt_signal2D(axes, transpose, tmp_path):
    s = Signal2D(arange(20).reshape((4, 5)))
    fname = tmp_path / "test.txt"
    savetxt(s, fname, axes=axes, transpose=transpose)
    s2 = loadtxt(fname)
    if axes:
        if transpose:
            assert_array_equal(s.axes_manager[1].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[0].axis, s2[1:, 0])
            assert_array_equal(s.data.T, s2[1:, 1:])
        else:
            assert_array_equal(s.axes_manager[0].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[1].axis, s2[1:, 0])
            assert_array_equal(s.data, s2[1:, 1:])
    else:
        if transpose:
            assert_array_equal(s.data.T, s2)
        else:
            assert_array_equal(s.data, s2)


@mark.parametrize("axes", (True, False))
@mark.parametrize("transpose", (True, False))
def test_savetxt_navigate2D(axes, transpose, tmp_path):
    s = Signal2D(arange(20).reshape((4, 5)))
    s.axes_manager[0].navigate = True
    s.axes_manager[1].navigate = True
    fname = tmp_path / "test.txt"
    savetxt(s, fname, axes=axes, transpose=transpose)
    s2 = loadtxt(fname)
    if axes:
        if transpose:
            assert_array_equal(s.axes_manager[1].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[0].axis, s2[1:, 0])
            assert_array_equal(s.data.T, s2[1:, 1:])
        else:
            assert_array_equal(s.axes_manager[0].axis, s2[0, 1:])
            assert_array_equal(s.axes_manager[1].axis, s2[1:, 0])
            assert_array_equal(s.data, s2[1:, 1:])
    else:
        if transpose:
            assert_array_equal(s.data.T, s2)
        else:
            assert_array_equal(s.data, s2)


def test_savetxt_dimension_error(tmp_path):
    s = LumiSpectrum(arange(60).reshape((3, 4, 5)))
    fname = tmp_path / "test.txt"
    with raises(NotImplementedError):
        s.savetxt(fname)


def test_to_array_dimension_error():
    s = LumiSpectrum(arange(60).reshape((3, 4, 5)))
    with raises(NotImplementedError):
        s.to_array()



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
Signal class for cathodoluminescence spectral data
--------------------------------------------------
"""

import numpy as np
from warnings import warn

from hyperspy._signals.lazy import LazySignal

from lumispy.signals import LumiSpectrum


class CLSpectrum(LumiSpectrum):
    """**General 1D cathodoluminescence signal class.**"""

    _signal_type = "CL"
    _signal_dimension = 1

    def _make_signal_mask(self, luminescence_roi):
        """Creates a mask from the peak position and peak widths of the
        luminescence spectrum.

        Parameters
        ----------
        luminescence_roi: array
            In the form of an array of pairwise elements
            [[peak1_x, peak1_width], [peak2_x, peak2_width],...].

        Returns
        -------
        array
            A `signal_mask`.
        """

        ax = self.axes_manager.signal_axes[0].axis
        signal_mask = np.ones(np.shape(ax))

        if len(np.shape(luminescence_roi)) == 1:
            luminescence_roi = np.array([luminescence_roi])

        for p in luminescence_roi:
            x, w = p
            x_min = x - w / 2
            x_max = x + w / 2
            index_min = np.abs(ax - x_min).argmin()
            index_max = np.abs(ax - x_max).argmin()
            signal_mask[index_min : index_max + 1] *= 0

        return np.invert(signal_mask.astype("bool"))

    def remove_spikes(
        self,
        threshold="auto",
        show_diagnosis_histogram=False,
        inplace=False,
        luminescence_roi=None,
        signal_mask=None,
        add_noise=False,
        navigation_mask=None,
        interactive=False,
        **kwargs,
    ):
        if luminescence_roi is not None and signal_mask is not None:
            raise AttributeError(
                "Only either `luminescence_roi` or the `signal_mask` can be an input."
            )

        if luminescence_roi is not None and signal_mask is None:
            signal_mask = self._make_signal_mask(luminescence_roi)

        if show_diagnosis_histogram:
            self.spikes_diagnosis(
                navigation_mask=navigation_mask, signal_mask=signal_mask, **kwargs
            )
        if inplace:
            signal = self
        else:
            signal = self.deepcopy()

        spikes_removal = signal.spikes_removal_tool(
            signal_mask=signal_mask,
            navigation_mask=navigation_mask,
            threshold=threshold,
            interactive=interactive,
            add_noise=add_noise,
            **kwargs,
        )

        if threshold == "auto":
            warn(
                "Threshold value: {:.2f}".format(spikes_removal.threshold), UserWarning
            )

        if inplace:
            return None
        else:
            return signal

    REMOVE_SPIKES_DOCSTRINGS = """HyperSpy-based spike removal tool adapted to
        LumiSpy to run non-interactively and without noise addition by default.
        %s
        
        Other Parameters
        ----------------
        show_diagnosis_histogram: bool
            Plot or not the derivative histogram to show the magnitude of the spikes present.
        inplace: bool
            If False, a new signal object is created and returned. If True, the original signal object is modified.
        luminescence_roi: array
            The peak position and peak widths of the peaks in the luminescence spectrum.
            In the form of an array of pairwise elements [[peak1_x, peak1_width], [peak2_x, peak2_width],...]
            in the units of the signal axis. It creates a signal_mask protecting the peak regions.
            To be used instead of `signal_mask`.

        Returns
        -------
        None or CLSpectrum
            Depends on inplace, returns or overwrites the CLSpectrum after spike removal.
        """
    remove_spikes.__doc__ = REMOVE_SPIKES_DOCSTRINGS % (
        LumiSpectrum.spikes_removal_tool.__doc__
    )


class LazyCLSpectrum(LazySignal, CLSpectrum):
    """**General lazy 1D cathodoluminescence signal class.**"""

    _lazy = True


"""SEM specific signal class for Cathodoluminescence spectral data.
"""


class CLSEMSpectrum(CLSpectrum):
    """**1D scanning electron microscopy cathodoluminescence signal class.**"""

    _signal_type = "CL_SEM"

    def correct_grating_shift(
        self, cal_factor_x_axis, corr_factor_grating, sem_magnification, **kwargs
    ):
        """Applies shift caused by the grating offset wrt the scanning centre.
        Authorship: Gunnar Kusch (gk419@cam.ac.uk)

        Parameters
        ----------
        cal_factor_x_axis
            The navigation correction factor.
        corr_factor_grating
            The grating correction factor.
        sem_magnification
            The SEM (real) magnification value.
            For the Attolight original metadata, take the `SEM.Real_Magnification` value
        kwargs
            The parameters passed to `hyperspy.align1D()` function like:
            interpolation_method ('linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic')
            parallel: Bool
            crop, expand, fill_value ...

        """

        # Don't correct for grating shift if this is already corrected
        if self.metadata.get_item("Signal.grating_corrected") is True:
            raise RuntimeError("The grating shift has already been corrected.")
        else:
            # Get all relevant parameters
            (nx, ny) = self.axes_manager.navigation_shape[:2]
            fov = sem_magnification

            # Correction of the Wavelength Shift along the X-Axis
            calax = cal_factor_x_axis / (fov * nx)
            # (Total Variation, Channels, Step)
            garray = np.arange(
                (-corr_factor_grating / 2) * calax * 1000 * nx,
                (corr_factor_grating / 2) * calax * 1000 * nx,
                corr_factor_grating * calax * 1000,
            )
            barray = np.full((ny, nx), garray)

            self.shift1D(barray, **kwargs)

            # Store modification in metadata
            self.metadata.set_item("Signal.grating_corrected", True)


class LazyCLSEMSpectrum(LazySignal, CLSEMSpectrum):
    """**Lazy 1D scanning electron microscopy cathodoluminescence signal class.**"""

    _lazy = True


"""STEM specific signal class for Cathodoluminescence spectral data.
"""


class CLSTEMSpectrum(CLSpectrum):
    """**1D scanning transmission electron microscopy cathodoluminescence signal class.**"""

    _signal_type = "CL_STEM"


class LazyCLSTEMSpectrum(LazySignal, CLSTEMSpectrum):
    """**Lazy 1D scanning transmission electron microscopy cathodoluminescence signal class.**"""

    _lazy = True



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
Signal class for luminescence data (BaseSignal class)
-----------------------------------------------------
"""

import numpy as np
from warnings import warn

from lumispy import nm2invcm
from lumispy.utils import (
    axis2eV,
    data2eV,
    var2eV,
    axis2invcm,
    data2invcm,
    var2invcm,
)


class CommonLumi:
    """**General luminescence signal class (dimensionless)**"""

    def crop_edges(self, crop_px):
        """Crop the amount of pixels from the four edges of the scanning
        region, from out the edges inwards.

        Parameters
        ----------
        crop_px : int
            Amount of pixels to be cropped on each side individually.

        Returns
        -------
        signal_cropped : CommonLuminescence
            A smaller cropped CL signal object. If inplace is True, the original
            object is modified and no LumiSpectrum is returned.
        """

        width = self.axes_manager.shape[0]
        height = self.axes_manager.shape[1]

        if crop_px * 2 > width or crop_px * 2 > height:
            raise ValueError(
                "The pixels to be cropped cannot be larger than half the width or the length!"
            )
        else:
            signal_cropped = self.inav[
                crop_px + 1 : width - crop_px + 1, crop_px + 1 : height - crop_px + 1
            ]

        # Store transformation in metadata (or update the value if already previously transformed)

        try:
            signal_cropped.metadata.Signal.cropped_edges += crop_px
        except AttributeError:
            signal_cropped.metadata.set_item("Signal.cropped_edges", crop_px)

        return signal_cropped

    def remove_negative(self, basevalue=1, inplace=False):
        """Sets all negative values to 'basevalue', e.g. for logarithmic scale
        plots.

        Parameters
        ----------
        basevalue : float
            Value by which negative values are replaced (default = 1).
        inplace : boolean
            If `False` (default), a new signal object is created and returned.
            Otherwise, the operation is performed on the existing signal object.

        Notes
        -----
        Sets `metadata.Signal.negative_removed` to `True`.
        """
        if inplace:
            s = self
        else:
            s = self.deepcopy()
        s.data[self.data < 0] = basevalue
        s.metadata.Signal.negative_removed = True
        if not inplace:
            return s

    def scale_by_exposure(self, integration_time=None, inplace=False, **kwargs):
        """Scale data in spectrum by integration time / exposure,
        (e.g. convert counts to counts/s).

        Parameters
        ----------
        integration_time : float
            Integration time (exposure) in s. If not given, the function tries to
            use the 'metadata.Acqusition_instrument.Detector.integration_time'
            field or alternatively find any 'integration_time', 'exposure' or
            'dwell_time' fields in the `original_metadata`.
        inplace : boolean
            If `False` (default), a new signal object is created and returned.
            If `True`, the operation is performed on the existing signal object.

        Notes
        -----
        Sets `metadata.Signal.scaled` to `True`. If intensity units is 'counts',
        replaces them by 'counts/s'.

        .. deprecated:: 0.2
          The `exposure` argument was renamed `integration_time`, and it will
          be removed in LumiSpy 1.0.
        """
        # Check metadata tags that would prevent scaling
        if self.metadata.Signal.get_item("normalized"):
            raise AttributeError("Data was normalized and cannot be scaled.")
        elif self.metadata.Signal.get_item("scaled") or self.metadata.Signal.get_item(
            "quantity"
        ) == ("Intensity (counts/s)" or "Intensity (Counts/s)"):
            raise AttributeError("Data was already scaled.")

        # Make sure integration_time is given or contained in metadata
        if integration_time is None:
            if "exposure" in kwargs:
                integration_time = kwargs["exposure"]
                warn(
                    "The `exposure` argument was renamed `integration_time` "
                    "and it will be removed in LumiSpy 1.0.",
                    DeprecationWarning,
                )
            elif self.metadata.has_item(
                "Acquisition_instrument.Detector.integration_time"
            ):
                integration_time = float(
                    self.metadata.get_item(
                        "Acquisition_instrument.Detector.integration_time"
                    )
                )
            else:
                raise AttributeError(
                    "Integration time (exposure) not given and it is not "
                    "included in the metadata."
                )
        if inplace:
            s = self
        else:
            s = self.deepcopy()
        s.data = s.data / integration_time
        s.metadata.Signal.scaled = True
        if s.metadata.get_item("Signal.quantity") == "Intensity (Counts)":
            s.metadata.Signal.quantity = "Intensity (Counts/s)"
            print(s.metadata.Signal.quantity)
        if s.metadata.get_item("Signal.quantity") == "Intensity (counts)":
            s.metadata.Signal.quantity = "Intensity (counts/s)"
        if not inplace:
            return s

    def normalize(self, pos=float("nan"), element_wise=False, inplace=False):
        """Normalizes data to value at `pos` along signal axis, defaults to
        maximum value.

        Can be helpful for e.g. plotting, but does not make sense to use
        on signals that will be used as input for further calculations!

        Parameters
        ----------
        pos : float, int
            If 'nan' (default), spectra are normalized to the maximum.
            If `float`, position along signal axis in calibrated units at which
            to normalize the spectra.
            If `int`, index along signal axis at which to normalize the spectra.
        element_wise: boolean
            If `False` (default), a spectrum image is normalized by a common factor.
            If `True`, each spectrum is normalized individually.
        inplace : boolean
            If `False` (default), a new signal object is created and returned.
            If `True`, the operation is performed on the existing signal object.

        Notes
        -----
        Sets `metadata.Signal.normalized` to `True`. If
        `metadata.Signal.quantity` contains the word 'Intensity', replaces this
        field with 'Normalized intensity'.
        """
        if self.metadata.Signal.get_item("normalized"):
            warn(
                "Data was already normalized previously. Depending on the "
                "previous parameters this function might not yield the "
                "expected result.",
                UserWarning,
            )
        if inplace:
            s = self
        else:
            s = self.deepcopy()
        # normalize on maximum
        if np.isnan(pos):
            if element_wise:
                s = s / s.max(axis=-1)
            else:
                s.data = s.data / s.max(axis=-1).max().data
        # normalize on given position along signal axis
        else:
            if element_wise:
                s = s / s.isig[pos]
            else:
                s.data = s.data / s.isig[pos].max().data
        s.metadata.Signal.normalized = True
        if s.metadata.get_item("Signal.quantity") is not None:
            if s.metadata.Signal.quantity.find("Intensity") != -1:
                s.metadata.Signal.quantity = "Normalized intensity"
        if not inplace:
            return s

    def _reset_variance_linear_model(self):
        """Resets the variance linear model parameters to their default values,
        as they are not applicable any longer after a Jacobian transformation.
        """
        if (
            (
                self.metadata.has_item(
                    "Signal.Noise_properties.Variance_linear_model.gain_factor"
                )
                and self.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor
                != 1
            )
            or (
                self.metadata.has_item(
                    "Signal.Noise_properties.Variance_linear_model.gain_offset"
                )
                and self.metadata.Signal.Noise_properties.Variance_linear_model.gain_offset
                != 0
            )
            or (
                self.metadata.has_item(
                    "Signal.Noise_properties.Variance_linear_model.correlation_factor"
                )
                and self.metadata.Signal.Noise_properties.Variance_linear_model.correlation_factor
                != 1
            )
        ):
            self.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor = 1
            self.metadata.Signal.Noise_properties.Variance_linear_model.gain_offset = 0
            self.metadata.Signal.Noise_properties.Variance_linear_model.correlation_factor = (
                1
            )
            warn(
                "Following the Jacobian transformation, the parameters of the "
                "`Variance_linear_model` are reset to their default values "
                "(gain_factor=1, gain_offset=0, correlation_factor=1).",
                UserWarning,
            )

    def _convert_data(self, newaxis, factor, inplace, jacobian, data2, var2):
        """Utility function to perform the data and variance conversion for
        signal unit transformations.
        """
        # convert data
        if jacobian:
            s2data = np.copy(
                data2(
                    self.isig[::-1].data,
                    factor,
                    newaxis.axis,
                    self.axes_manager.signal_axes[0],
                )
            )
        else:
            s2data = np.copy(self.isig[::-1].data)

        # inplace conversion
        if inplace:
            s2 = self
            s2.data = s2data
        # create new signal object with correct data, axes, metadata
        else:
            s2 = self._deepcopy_with_new_data(
                s2data,
                copy_variance=(not jacobian),
                copy_learning_results=False,
            )
        # convert axis
        oldaxis = self.axes_manager.signal_axes[0]
        axind = self.axes_manager.signal_axes[0].index_in_axes_manager
        # workaround for bug in set_axis that changes wrong axis
        if self.axes_manager.signal_dimension == 2:
            axind += 1
        s2.axes_manager.set_axis(newaxis, axind)
        # convert variance
        if self.metadata.has_item("Signal.Noise_properties.variance"):
            var = self.get_noise_variance()
            if jacobian:
                # if variance is a numeric value, cast into signal object
                if isinstance(var, (float, int)):
                    var = self._deepcopy_with_new_data(
                        np.ones(self.data.shape) * var,
                        copy_variance=False,
                        copy_learning_results=False,
                    )
                s2var = s2._deepcopy_with_new_data(
                    var2(
                        var.isig[::-1].data,
                        factor,
                        newaxis.axis,
                        oldaxis,
                    ),
                    copy_variance=False,
                    copy_learning_results=False,
                )
                s2.set_noise_variance(s2var)
                s2._reset_variance_linear_model()
            else:
                # variance left unchanged, if it is a number and Jacobian not performed
                if isinstance(var, (float, int)):
                    if not inplace:
                        s2.set_noise_variance(self.get_noise_variance())
                else:
                    s2.set_noise_variance(
                        s2._deepcopy_with_new_data(
                            var.isig[::-1].data,
                            copy_variance=False,
                            copy_learning_results=False,
                        )
                    )
        if not inplace:
            return s2
        else:
            return None

    def to_eV(self, inplace=True, jacobian=True):
        """Converts signal axis of 1D signal to non-linear energy axis (eV)
        using wavelength dependent refractive index of air. Assumes wavelength
        in units of nm unless the axis units are specifically set to µm.

        The intensity is converted from counts/nm (counts/µm) to counts/meV by
        doing a Jacobian transformation, see e.g. Mooney and Kambhampati, J.
        Phys. Chem. Lett. 4, 3316 (2013), doi:10.1021/jz401508t, which ensures
        that integrated signals are correct also in the energy domain. If the
        variance of the signal is known, i.e.
        `metadata.Signal.Noise_properties.variance` is a signal representing
        the variance, a squared renormalization of the variance is performed.
        Note that if the variance is a number (not a signal instance), it is
        converted to a signal if the Jacobian transformation is performed

        Parameters
        ----------
        inplace : boolean
            If `False`, a new signal object is created and returned. Otherwise
            (default) the operation is performed on the existing signal object.
        jacobian : boolean
            The default is to do the Jacobian transformation (recommended at
            least for luminescence signals), but the transformation can be
            suppressed by setting this option to `False`.

        Examples
        --------
        >>> import numpy as np
        >>> from lumispy import LumiSpectrum
        >>> S1 = LumiSpectrum(np.ones(20), DataAxis(axis = np.arange(200,400,10)), ))
        >>> S1.to_eV()
        """

        evaxis, factor = axis2eV(self.axes_manager.signal_axes[0])

        return self._convert_data(evaxis, factor, inplace, jacobian, data2eV, var2eV)

    TO_INVCM_DOCSTRING = """
        The intensity is converted from counts/nm (counts/µm) to counts/cm^-1
        by doing a Jacobian transformation, see e.g. Mooney and Kambhampati,
        J. Phys. Chem. Lett. 4, 3316 (2013), doi:10.1021/jz401508t, which
        ensures that integrated signals are correct also in the wavenumber
        domain. If the variance of the signal is known, i.e.
        `metadata.Signal.Noise_properties.variance` is a signal representing the
        variance, a squared renormalization of the variance is performed.
        Note that if the variance is a number (not a signal instance), it is
        converted to a signal if the Jacobian transformation is performed

        Parameters
        ----------
        inplace : boolean
            If `False`, a new signal object is created and returned. Otherwise
            (default) the operation is performed on the existing signal object.
        """

    TO_INVCM_EXAMPLE = """
        Examples
        --------
        >>> import numpy as np
        >>> from lumispy import LumiSpectrum
        >>> S1 = LumiSpectrum(np.ones(20), DataAxis(axis = np.arange(200,400,10)), ))
        >>> S1.to_invcm()
        """

    def to_invcm(self, inplace=True, jacobian=True):
        """Converts signal axis of 1D signal to non-linear wavenumber axis
        (cm^-1). Assumes wavelength in units of nm unless the axis units are
        specifically set to µm.
        %s
        jacobian : boolean
            The default is to do the Jacobian transformation (recommended at
            least for luminescence signals), but the transformation can be
            suppressed by setting this option to `False`.
        %s
        """

        invcmaxis, factor = axis2invcm(self.axes_manager.signal_axes[0])

        return self._convert_data(
            invcmaxis, factor, inplace, jacobian, data2invcm, var2invcm
        )

    to_invcm.__doc__ %= (TO_INVCM_DOCSTRING, TO_INVCM_EXAMPLE)

    TO_INVCMREL_EXAMPLE = """
        Examples
        --------
        >>> import numpy as np
        >>> from lumispy import LumiSpectrum
        >>> S1 = LumiSpectrum(np.ones(20), DataAxis(axis = np.arange(200,400,10)), ))
        >>> S1.to_invcm(laser=325)
    """

    def to_invcm_relative(self, laser=None, inplace=True, jacobian=False):
        """Converts signal axis of 1D signal to non-linear wavenumber axis
        (cm^-1) relative to the exciting laser wavelength (Raman/Stokes
        shift). Assumes wavelength in units of nm unless the axis units are
        specifically set to µm.
        %s
        laser: float or None
            Laser wavelength in the same units as the signal axis. If None
            (default), checks if it is stored in
            `metadata.Acquisition_instrument.Laser.wavelength`.
        jacobian : boolean
            The default is not to do the Jacobian transformation for Raman
            shifts, but the transformation can be activated by setting this
            option to `True`.
        %s
        """

        # check if laser wavelength is available
        if laser is None:
            if not self.metadata.has_item("Acquisition_instrument.Laser.wavelength"):
                raise AttributeError(
                    "Laser wavelength is neither given in the metadata nor passed"
                    " to the function."
                )
            else:
                laser = self.metadata.get_item(
                    "Acquisition_instrument.Laser.wavelength"
                )
        # check if laser units make sense in respect to signal units
        if (self.axes_manager.signal_axes[0].units == "µm" and laser > 10) or (
            self.axes_manager.signal_axes[0].units == "nm" and laser < 100
        ):
            raise AttributeError(
                "Laser wavelength units do not seem to match the signal units."
            )

        invcmaxis, factor = axis2invcm(self.axes_manager.signal_axes[0])

        # convert to relative wavenumber scale
        if self.axes_manager.signal_axes[0].units == "µm":
            invcmlaser = nm2invcm(1000 * laser)
        else:
            invcmlaser = nm2invcm(laser)
        absaxis = invcmaxis.axis[::-1]
        invcmaxis.axis = invcmlaser - absaxis
        invcmaxis.name = "Raman Shift"

        # replace signal axis after conversion
        # conversion (using absolute scale for Jacobian)
        if inplace:
            self.to_invcm(inplace=inplace, jacobian=jacobian)
            s2 = self
        else:
            s2 = self.to_invcm(inplace=inplace, jacobian=jacobian)
        # replace axis
        axind = self.axes_manager.signal_axes[0].index_in_axes_manager
        # workaround for bug in set_axis that changes wrong axis
        if self.axes_manager.signal_dimension == 2:
            axind += 1
        s2.axes_manager.set_axis(invcmaxis, axind)
        s2.data = s2.isig[::-1].data
        # replace variance axis
        if s2.metadata.has_item("Signal.Noise_properties.variance") and not isinstance(
            s2.get_noise_variance(), (float, int)
        ):
            axind = self.axes_manager.signal_axes[0].index_in_axes_manager
            # workaround for bug in set_axis that changes wrong axis
            if self.axes_manager.signal_dimension == 2:
                axind += 1
            s2.metadata.Signal.Noise_properties.variance.axes_manager.set_axis(
                invcmaxis,
                axind,
            )
            s2.metadata.Signal.Noise_properties.variance.data = (
                s2.metadata.Signal.Noise_properties.variance.isig[::-1].data
            )
        if not inplace:
            return s2
        else:
            return None

    to_invcm_relative.__doc__ %= (
        TO_INVCM_DOCSTRING,
        TO_INVCMREL_EXAMPLE,
    )

    # Alias Method Name
    to_raman_shift = to_invcm_relative



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
Signal class for luminescence transient data (2D)
-------------------------------------------------
"""

import pint

from hyperspy.signals import Signal1D, Signal2D
from hyperspy._signals.lazy import LazySignal
from hyperspy.docstrings.signal import OPTIMIZE_ARG

from lumispy.signals import LumiSpectrum, LumiTransient
from lumispy.signals.common_luminescence import CommonLumi
from lumispy.signals.common_transient import CommonTransient


class TransientSpectrumCasting(Signal1D, CommonLumi, CommonTransient):
    """**Hidden signal class**
    1D version of ``TransientSpectrum`` signal class for casting
    ``LumiTransientSpectrum` to either ``Luminescence`` or ``Transient``
    when the signal dimensionality is reduced.

    Example:
    --------

    >>> s = LumiTransientSpectrum(np.random.random((10, 10, 10, 10))) * 2
    >>> s.axes_manager.signal_axes[-1].units = 'ps'
    >>> s.axes_manager.signal_axes[0].units = 'nm'
    >>> s.sum(axis=-1)
    >>> s
    <LumiSpectrum, title: , dimensions: (10, 10|10)>
    """

    _signal_type = "TransientSpectrum"
    _signal_dimension = 1

    def __init__(self, *args, **kwargs):
        ureg = pint.UnitRegistry()
        if (
            hasattr(self, "axes_manager")
            and ureg(self.axes_manager.signal_axes[-1].units).dimensionality
            == ureg("s").dimensionality
        ):
            self.set_signal_type("Transient")
        else:
            self.set_signal_type("Luminescence")


class LumiTransientSpectrum(Signal2D, CommonLumi, CommonTransient):
    """**2D luminescence signal class (spectrum+transient/time resolved dimensions)**"""

    _signal_type = "TransientSpectrum"
    _signal_dimension = 2

    def spec2nav(self, optimize=True):
        """Return the streak image as signal with the spectral axis as navigation
        axis and the time axis as signal axis. For efficient iteration over
        transients as a function of the spectral positions (e.g. for fitting
        transients). By default, the method ensures that the data is stored optimally,
        hence often making a copy of the data.

        Parameters
        ----------
        %s

        Returns
        -------
        signal : LumiSpectrum
            A signal of type ``LumiTransient``.

        See Also
        --------
        lumispy.signals.LumiTransientSpectrum.time2nav
        hyperspy.api.signals.BaseSignal.transpose
        """
        s = self.transpose(signal_axes=[-1], optimize=optimize)
        return s

    spec2nav.__doc__ %= (OPTIMIZE_ARG,)

    def time2nav(self, optimize=True):
        """Return the streak image as signal with the time axis as navigation
        axis and the spectral axis as signal axis. For efficient iteration over
        spectra as a function of time (e.g. for fitting spectra). By default, the
        method ensures that the data is stored optimally, hence often making a copy
        of the data.

        Parameters
        ----------
        %s

        Returns
        -------
        signal : LumiSpectrum
            A signal of type ``LumiSpectrum``.

        See Also
        --------
        lumispy.signals.LumiTransientSpectrum.time2nav
        hyperspy.api.signals.BaseSignal.transpose
        """
        s = self.transpose(signal_axes=[-2], optimize=optimize)
        return s

    time2nav.__doc__ %= (OPTIMIZE_ARG,)


class LazyLumiTransientSpectrum(LazySignal, LumiTransientSpectrum):
    """**Lazy 2D luminescence signal class (spectral+transient/time resolved dimensions)**"""

    _lazy = True



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from .luminescence_spectrum import LumiSpectrum, LazyLumiSpectrum
from .cl_spectrum import CLSpectrum, LazyCLSpectrum
from .cl_spectrum import CLSEMSpectrum, LazyCLSEMSpectrum
from .cl_spectrum import CLSTEMSpectrum, LazyCLSTEMSpectrum
from .pl_spectrum import PLSpectrum, LazyPLSpectrum
from .el_spectrum import ELSpectrum, LazyELSpectrum
from .luminescence_transient import LumiTransient, LazyLumiTransient
from .luminescence_transientspec import LumiTransientSpectrum, LazyLumiTransientSpectrum


__all__ = [
    "LumiSpectrum",
    "LazyLumiSpectrum",
    "CLSpectrum",
    "LazyCLSpectrum",
    "CLSEMSpectrum",
    "LazyCLSEMSpectrum",
    "CLSTEMSpectrum",
    "LazyCLSTEMSpectrum",
    "PLSpectrum",
    "LazyPLSpectrum",
    "ELSpectrum",
    "LazyELSpectrum",
    "LumiTransient",
    "LazyLumiTransient",
    "LumiTransientSpectrum",
    "LazyLumiTransientSpectrum",
]



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""Signal class for Luminescence spectral data (1D).
"""
import numpy as np
from warnings import warn

from hyperspy.signals import Signal1D
from hyperspy._signals.lazy import LazySignal
from traits.api import Undefined

from lumispy.signals.common_luminescence import CommonLumi
from lumispy import to_array, savetxt
from lumispy.utils import solve_grating_equation
from lumispy.utils.axes import GRATING_EQUATION_DOCSTRING_PARAMETERS
from lumispy.utils.signals import com
from lumispy.utils.io import (
    SAVETXT_DOCSTRING,
    SAVETXT_PARAMETERS,
    TOARRAY_DOCSTRING,
    TOARRAY_PARAMETERS,
)


class LumiSpectrum(Signal1D, CommonLumi):
    """**General 1D luminescence signal class.**"""

    _signal_dimension = 1
    _signal_type = "Luminescence"

    def __init(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def remove_background_from_file(self, background=None, inplace=False, **kwargs):
        """Subtract the background to the signal in all navigation axes. If no
        background file is passed as argument, the `remove_background()` from
        HyperSpy is called with the GUI.

        Parameters
        ----------
        background : array shape (2, n) or Signal1D
            An array containing the background x-axis and the intensity values
            [[xs],[ys]] or a Signal1D object. If the x-axis values do not match
            the signal_axes, then interpolation is done before subtraction. If
            only the intensity values are provided, [ys], the functions assumes
            no interpolation needed.
        inplace : boolean
            If False, it returns a new object with the transformation. If True,
            the original object is transformed, returning no object.

        Returns
        -------
        signal : LumiSpectrum
            A background subtracted signal.

        Notes
        -----
        This function does not work with non-uniform axes.
        """
        warn(
            "The use of `remove_background_from_file` is deprecated and will "
            "be removed in LumiSpy 1.0. Please use `remove_background_signal` "
            "from the Signal1D class.",
            DeprecationWarning,
            stacklevel=2,
        )
        if hasattr(self.metadata.Signal, "background_subtracted"):
            if self.metadata.Signal.background_subtracted is True:
                raise RecursionError(
                    "You have already removed background once. If you need to "
                    "remove it again, set the "
                    "s.metadata.Signal.background_subtracted to False."
                )
        elif background is None:  # pragma: no cover
            warn(
                "Using the Hyperspy specfic `remove_background` function. "
                "Use `s.remove_background()` instead.",
                category=SyntaxWarning,
            )
            self.remove_background(**kwargs)
        else:
            signal_x = self.axes_manager.signal_axes[0].axis

            if hasattr(background, "axes_manager"):  # Check if Hyperspy-like object
                x = background.axes_manager.signal_axes[0].axis
                y = background.data
                background = [x, y]

            background_xy = background

            if np.shape(background_xy)[0] == 1:
                bkg_x = signal_x
                bkg_y = background_xy[0]
            elif np.shape(background_xy)[0] == 2:
                try:
                    bkg_x = background_xy[0]
                    bkg_y = background_xy[1]
                except IndexError:  # pragma: no cover
                    raise AttributeError(
                        "Please provide a background file containing both the x and y axis."
                    )
            else:
                raise AttributeError(
                    "Please, provide a background of shape (2, n) or (n,)"
                )

            if not np.array_equal(bkg_x, signal_x):
                # Interpolation needed
                bkg_y = np.interp(signal_x, bkg_x, bkg_y)

            if not inplace:
                self_subtracted = self.map(
                    lambda s, bkg: s - bkg, bkg=bkg_y, inplace=False
                )
                self_subtracted.metadata.set_item("Signal.background_subtracted", True)
                self_subtracted.metadata.set_item("Signal.background", bkg_y)
                return self_subtracted
            else:
                self.metadata.set_item("Signal.background_subtracted", True)
                self.metadata.set_item("Signal.background", bkg_y)
                return self.map(lambda s, bkg: s - bkg, bkg=bkg_y, inplace=True)

    SAVETXT_EXAMPLE = """
    Examples
    --------
    >>> import lumispy as lum
    >>> import numpy as np
    ...
    >>> # Spectrum:
    >>> S = lum.signals.LumiSpectrum(np.arange(5))
    >>> S.savetxt('spectrum.txt')
    0.00000	0.00000
    1.00000	1.00000
    2.00000	2.00000
    3.00000	3.00000
    4.00000	4.00000
    ...
    >>> # Linescan:
    >>> L = lum.signals.LumiSpectrum(np.arange(25).reshape((5,5)))
    >>> L.savetxt('linescan.txt')
    0.00000	0.00000	1.00000	2.00000	3.00000	4.00000
    0.00000	0.00000	5.00000	10.00000	15.00000	20.00000
    1.00000	1.00000	6.00000	11.00000	16.00000	21.00000
    2.00000	2.00000	7.00000	12.00000	17.00000	22.00000
    3.00000	3.00000	8.00000	13.00000	18.00000	23.00000
    4.00000	4.00000	9.00000	14.00000	19.00000	24.00000
    """

    def savetxt(
        self,
        filename,
        fmt="%.5f",
        delimiter="\t",
        axes=True,
        transpose=False,
        **kwargs,
    ):
        """Writes luminescence spectrum object to simple text file.
        %s
        %s
        %s
        """
        savetxt(self, filename, fmt, delimiter, axes, transpose, **kwargs)

    savetxt.__doc__ %= (
        SAVETXT_DOCSTRING,
        SAVETXT_PARAMETERS,
        SAVETXT_EXAMPLE,
    )

    TOARRAY_EXAMPLE = """
    Notes
    -----
    The output of this function can be used to convert a signal object to a
    pandas dataframe, e.g. using `df = pd.Dataframe(S.to_array())`.

    Examples
    --------
    >>> import lumispy as lum
    >>> import numpy as np
    ...
    >>> # Spectrum:
    >>> S = lum.signals.LumiSpectrum(np.arange(5))
    >>> S.to_array()
    array([[0., 0.],
          [1., 1.],
          [2., 2.],
          [3., 3.],
          [4., 4.]])
    ...
    >>> # Linescan:
    >>> L = lum.signals.LumiSpectrum(np.arange(25).reshape((5,5)))
    >>> L.to_array()
    array([[ 0.,  0.,  1.,  2.,  3.,  4.],
          [ 0.,  0.,  1.,  2.,  3.,  4.],
          [ 1.,  5.,  6.,  7.,  8.,  9.],
          [ 2., 10., 11., 12., 13., 14.],
          [ 3., 15., 16., 17., 18., 19.],
          [ 4., 20., 21., 22., 23., 24.]])
    """

    def to_array(self, axes=True, transpose=False):
        """Returns luminescence spectrum object as numpy array (optionally
        including the axes).
            %s
            %s
        %s
        """
        return to_array(self, axes, transpose)

    to_array.__doc__ %= (
        TOARRAY_DOCSTRING,
        TOARRAY_PARAMETERS,
        TOARRAY_EXAMPLE,
    )

    def px_to_nm_grating_solver(
        self,
        gamma_deg,
        deviation_angle_deg,
        focal_length_mm,
        ccd_width_mm,
        grating_central_wavelength_nm,
        grating_density_gr_mm,
        inplace=False,
    ):
        """Converts signal axis of 1D signal (in pixels) to wavelength, solving
        the grating equation. See `lumispy.axes.solve_grating_equation` for
        more details.

        Parameters
        ----------
        %s
            inplace : bool
                If False, it returns a new object with the transformation. If
                True, the original object is transformed, returning no object.

        Returns
        -------
        signal : LumiSpectrum
            A signal with calibrated wavelength units.

        Examples
        --------
        >>> s = LumiSpectrum(np.ones(20),))
        >>> s.px_to_nm_grating_solver(*params, inplace=True)
        >>> s.axes_manager.signal_axes[0].units == 'nm'
        """

        nm_axis = solve_grating_equation(
            self.axes_manager.signal_axes[0],
            gamma_deg,
            deviation_angle_deg,
            focal_length_mm,
            ccd_width_mm,
            grating_central_wavelength_nm,
            grating_density_gr_mm,
        )

        if inplace:
            s = self
        else:
            s = self.deepcopy()

        s.axes_manager.set_axis(
            nm_axis,
            self.axes_manager.signal_axes[0].index_in_axes_manager,
        )

        return s

    px_to_nm_grating_solver.__doc__ %= GRATING_EQUATION_DOCSTRING_PARAMETERS.replace(
        "\n", "\n\t"
    )

    def centroid(self, signal_range=None, **kwargs):
        """
        Finds the centroid (center of mass) of a peak in the spectrum from
        the wavelength (or pixel number) and the intensity at each pixel
        value. It basically represents a "weighted average" of the peak.

        Notes
        -----
        This function only works for a single peak. If you have multiple
        peaks, slice the signal beforehand or use the signal_range parameter.

        TODO: Implement this function for multiple peaks (with the npeaks
        parameter) by finding the top 2 peaks from mean spectrum and then
        returning a signal with 2 com.

        Parameters
        ----------
        signal_range : tuple of ints or floats, optional
            A tuple representing the indices of the signal axis (start index,
            end index) where the peak is located. If the tuple contains int,
            it slices on index. If the tuple contains float, it slices on
            signal units (default HyperSpy s.inav[:] functionality).

        kwargs : dictionary
            For the scipy.interpolate.interp1d function.

        Returns
        -------
        signal : Signal2D, BaseSignal
            Signal object containing the center of mass for every pixel. Depending
            on the dimensionality the type is Signal2D or a BaseSignal (for single
            spectrum).
        """
        if signal_range:
            if type(signal_range) != tuple:
                raise TypeError(
                    "The `signal_range` parameter must be a tuple of length 2."
                )
            if len(signal_range) != 2:
                raise ValueError(
                    f"The `signal_range` parameter must be a tuple of length 2. "
                    "You passed a tuple of length {len(signal_range)}."
                )

            s = self.isig[signal_range[0] : signal_range[1]]

        else:
            s = self

        signal_axis = s.axes_manager.signal_axes[0]
        center_of_mass = s.map(com, signal_axis=signal_axis, inplace=False)

        # Transfer axes metadata to title
        center_of_mass.metadata.General.title = f"Centroid map"
        if signal_axis.name not in (Undefined, ""):
            center_of_mass.metadata.General.title += f" of {signal_axis.name}"
        if signal_axis.units not in (Undefined, ""):
            center_of_mass.metadata.General.title += f" ({signal_axis.units})"
        if s.metadata.General.title not in (Undefined, ""):
            center_of_mass.metadata.General.title += f" for {s.metadata.General.title}"
        if center_of_mass.axes_manager.navigation_size > 0:
            center_of_mass = center_of_mass.transpose()
        return center_of_mass


class LazyLumiSpectrum(LazySignal, LumiSpectrum):
    """**General lazy 1D luminescence signal class.**"""

    _lazy = True



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import scipy.constants as c
from scipy.interpolate import interp1d
from inspect import getfullargspec
from copy import deepcopy
from warnings import warn

from hyperspy.axes import DataAxis, UniformDataAxis


#
# Functions needed for signal axis conversion
#


def _n_air(x):
    """Refractive index of air as a function of WL in nm.
    This analytical function is correct for the range 185-1700 nm.
    According to `E.R. Peck and K. Reeder. Dispersion of air,
    J. Opt. Soc. Am. 62, 958-962 (1972).`
    """
    wl = deepcopy(x)
    # Check for supported range
    if (np.min(wl) < 185) or (np.max(wl) > 1700):
        if np.size(wl) == 1:
            if wl < 185:
                wl = 185
            if wl > 1700:
                wl = 1700
        else:
            wl[wl < 185] = 185
            wl[wl > 1700] = 1700
        warn(
            "The wavelength range exceeds the interval of 185 to 1700 nm for "
            "which the exact refractive index of air is used. Beyond this "
            "range, the refractive index is kept constant.",
            UserWarning,
        )
    wl = wl / 1000
    return (
        1
        + 806051e-10
        + 2480990e-8 / (132274e-3 - 1 / wl**2)
        + 174557e-9 / (3932957e-5 - 1 / wl**2)
    )


def nm2eV(x):
    """Converts wavelength (nm) to energy (eV) using wavelength-dependent
    refractive index of air.
    """
    return 1e9 * c.h * c.c / (c.e * _n_air(x) * x)


def eV2nm(x):
    """Converts energy (eV) to wavelength (nm) using wavelength-dependent
    refractive index of air.
    """
    wl = 1239.5 / x  # approximate WL to obtain refractive index
    return 1e9 * c.h * c.c / (c.e * _n_air(wl) * x)


def axis2eV(ax0):
    """Converts given signal axis to energy scale (eV) using wavelength
    dependent refractive index of air. Assumes wavelength in units of nm unless the
    axis units are specifically set to µm.
    """
    if ax0.units == "eV":
        raise AttributeError("Signal unit is already eV.")
    # transform axis, invert direction
    if ax0.units == "µm":
        evaxis = nm2eV(1000 * ax0.axis)[::-1].astype("float")
        factor = 1e3  # correction factor for intensity
    else:
        evaxis = nm2eV(ax0.axis)[::-1].astype("float")
        factor = 1e6
    axis = DataAxis(axis=evaxis, name="Energy", units="eV", navigate=False)
    return axis, factor


def data2eV(data, factor, evaxis, ax0):
    """The intensity is converted from counts/nm (counts/µm) to counts/meV by
    doing a Jacobian transformation, see e.g. Mooney and Kambhampati, J. Phys.
    Chem. Lett. 4, 3316 (2013). Ensures that integrated signals are still
    correct.
    """
    if ax0.units == "µm":
        return (
            data
            * factor
            * c.h
            * c.c
            / (c.e * _n_air(1000 * ax0.axis)[::-1] * evaxis**2)
        )
    else:
        return data * factor * c.h * c.c / (c.e * _n_air(ax0.axis[::-1]) * evaxis**2)


def var2eV(variance, factor, evaxis, ax0):
    """The variance is converted doing a squared Jacobian renormalization to
    match with the transformation of the data.
    """
    if ax0.units == "µm":
        return (
            variance
            * (factor * c.h * c.c / (c.e * _n_air(1000 * ax0.axis)[::-1] * evaxis**2))
            ** 2
        )
    else:
        return (
            variance
            * (factor * c.h * c.c / (c.e * _n_air(ax0.axis[::-1]) * evaxis**2)) ** 2
        )


def nm2invcm(x):
    r"""Converts wavelength (nm) to wavenumber (cm$^{-1}$)."""
    return 1e7 / x


def invcm2nm(x):
    r"""Converts wavenumber (cm$^{-1}$) to wavelength (nm)."""
    return 1e7 / x


def axis2invcm(ax0):
    r"""Converts given signal axis to wavenumber scale (cm$^{-1}$). Assumes
    wavelength in units of nm unless the axis units are specifically set to µm.
    """
    if ax0.units == r"cm$^{-1}$":
        raise AttributeError(r"Signal unit is already cm$^{-1}$.")
    # transform axis, invert direction
    if ax0.units == "µm":
        invcmaxis = nm2invcm(1000 * ax0.axis)[::-1].astype("float")
        factor = 1e4  # correction factor for intensity
    else:
        invcmaxis = nm2invcm(ax0.axis)[::-1].astype("float")
        factor = 1e7
    axis = DataAxis(
        axis=invcmaxis, name="Wavenumber", units=r"cm$^{-1}$", navigate=False
    )
    return axis, factor


def data2invcm(data, factor, invcmaxis, ax0=None):
    r"""The intensity is converted from counts/nm (counts/µm) to
    counts/cm$^{-1}$ by doing a Jacobian transformation, see e.g. Mooney and
    Kambhampati, J. Phys. Chem. Lett. 4, 3316 (2013). Ensures that integrated
    signals are still correct.
    """
    return data * factor / (invcmaxis**2)


def var2invcm(variance, factor, invcmaxis, ax0=None):
    r"""The variance is converted doing a squared Jacobian renormalization to
    match with the transformation of the data.
    """
    return variance * (factor / (invcmaxis**2)) ** 2


#
# spectrum manipulation
#


def join_spectra(S, r=50, scale=True, average=False, kind="slinear"):
    """Takes list of Signal1D objects and returns a single object with all
    spectra joined. Joins spectra at the center of the overlapping range.
    Scales spectra by a factor determined as average over the range
    `center -/+ r` pixels. Works both for uniform and non-uniform axes
    (FunctionalDataAxis is converted into a non-uniform DataAxis).

    Parameters
    ----------
    S : list of Signal1D objects (with overlapping signal axes)
    r : int, optional
        Number of pixels left/right of center (default `50`) defining the range
        over which to determine the scaling factor, has to be less than half
        of the overlapping pixels. Change the size of `r` or use `average=True`
        if the function induces a step in the intensity.
    scale : boolean, optional
        If `True` (default), the later spectra in the list are scaled by a
        factor determined over `center -/+ r` pixels. If `False`, spectra are
        joined without scaling, which will likely induce a step unless
        `average=True`.
    average : boolean, optional
        If `True`, the contribution of the two signals is continuously graded
        within the range defined by `r` instead of joining at the center of
        the range (default).
    kind : str, optional
        Interpolation method (default 'slinear') to use when joining signals
        with a uniform signal axes. See `scipy.interpolate.interp1d` for
        options.

    Returns
    -------
    A new Signal1D object containing the joined spectra (properties are copied
    from first spectrum).

    Examples
    --------

    >>> s1 = hs.signals.Signal1D(np.ones(32))
    >>> s2 = hs.signals.Signal1D(np.ones(32)*2)
    >>> s2.axes_manager.signal_axes[0].offset = 25
    >>> lum.join_spectra([s1,s2],r=2)
    <Signal1D, title: , dimensions: (|57)>
    """

    # Test that spectra overlap
    for i in range(1, len(S)):
        if (
            S[i - 1].axes_manager.signal_axes[0].axis.max()
            < S[i].axes_manager.signal_axes[0].axis.min()
        ):
            raise ValueError("Signal axes not overlapping")
    # Make sure that r is of type int
    if not type(r) is int:
        r = int(r)

    # take first spectrum as basis
    S1 = S[0].deepcopy()
    axis = S1.axes_manager.signal_axes[0]
    for i in range(1, len(S)):  # join following spectra
        S2 = S[i].deepcopy()
        axis2 = S2.axes_manager.signal_axes[0]
        omax = axis.axis.max()  # define overlap range
        omin = axis2.axis.min()
        ocenter = (omax + omin) / 2  # center of overlap range
        # closest index to center of overlap first spectrum
        ind1 = axis.value2index(ocenter)
        # closest index to center of overlap second spectrum
        ind2 = axis2.value2index(ocenter)
        # Test that r is not too large
        if (axis.size - ind1 - 1) <= r:
            raise ValueError("`r` is too large")
        # calculate mean deviation over defined range ignoring nan/zero values
        init = np.empty(S2.isig[ind2 - r : ind2 + r].data.shape)
        init[:] = np.nan
        # Do scaling of following signals
        if scale:
            if (
                axis.axis[ind1 - r : ind1 + r] == axis2.axis[ind2 - r : ind2 + r]
            ).all():
                factor = np.nanmean(
                    np.ma.masked_invalid(
                        np.divide(
                            S1.isig[ind1 - r : ind1 + r].data,
                            S2.isig[ind2 - r : ind2 + r].data,
                            out=init,
                            where=S2.isig[ind2 - r : ind2 + r].data != 0,
                        )
                    ),
                    axis=-1,
                )
            else:  # interpolate to get factor at same positions
                ind2r1 = axis2.value2index(axis.axis[ind1 - r])
                ind2r2 = axis2.value2index(axis.axis[ind1 + r])
                f = interp1d(
                    axis2.axis[ind2r1 - 1 : ind2r2 + 1],
                    S2.isig[ind2r1 - 1 : ind2r2 + 1].data,
                    kind=kind,
                )
                factor = np.nanmean(
                    np.ma.masked_invalid(
                        np.divide(
                            S1.isig[ind1 - r : ind1 + r].data,
                            f(axis.axis[ind1 - r : ind1 + r]),
                            out=init,
                            where=S2.isig[ind2 - r : ind2 + r].data != 0,
                        )
                    ),
                    axis=-1,
                )
            if (factor < 0).any():
                raise ValueError(
                    "One of the signals has a negative mean"
                    " value in the overlapping range. Try to set"
                    " `scale=False` and `average=True`."
                )
            S2.data = (S2.data.T * factor).T  # scale 2nd spectrum by factor
        # Make sure the corresponding values are in correct order
        if axis.axis[ind1] >= axis2.axis[ind2]:
            ind2 += 1
            # for UniformDataAxis
        if (not "axis" in getfullargspec(DataAxis)[0]) or (
            axis.is_uniform and axis2.is_uniform
        ):
            # join axis vectors
            axis.size = axis.axis[: ind1 + 1].size + np.floor(
                (axis2.axis[-1] - axis.axis[ind1]) / axis.scale
            )
            # join data vectors interpolating to a common uniform axis
            if average:  # average over range
                if r == 1:
                    raise ValueError(
                        "Averaging can not be performed for r=1. "
                        "Set average=False or r>1."
                    )
                ind2r = axis2.value2index(axis.axis[ind1 - r])
                length = axis.axis[ind1 - r : ind1 + r].size
                grad = 1 / (length - 1)
                vect = np.arange(length)
                f = interp1d(
                    axis2.axis[ind2r - 1 :], S2.isig[ind2r - 1 :].data, kind=kind
                )
                S1.data = np.hstack(
                    (
                        S1.isig[: ind1 - r].data,
                        (1 - grad * vect) * S1.isig[ind1 - r : ind1 + r].data
                        + grad * vect * f(axis.axis[ind1 - r : ind1 + r]),
                        f(axis.axis[ind1 + r :]),
                    )
                )
            else:  # just join at center of overlap
                f = interp1d(axis2.axis[ind2:], S2.isig[ind2:].data, kind=kind)
                S1.data = np.hstack(
                    (S1.isig[: ind1 + 1].data, f(axis.axis[ind1 + 1 :]))
                )
        else:  # for DataAxis/FunctionalDataAxis (non uniform)
            # convert FunctionalDataAxes or UniformDataAxis to DataAxes
            if hasattr(axis, "_expression") or axis.is_uniform:
                axis.convert_to_non_uniform_axis()
            # 2nd axis does not need to be converted, because it contains axis vector
            # join axis vectors
            axis.axis = np.hstack((axis.axis[: ind1 + 1], axis2.axis[ind2:]))
            axis.size = axis.axis.size
            if average:  # average over range
                f1 = interp1d(
                    S[i - 1].axes_manager.signal_axes[0].axis[ind1 - 1 : ind1 + r + 1],
                    S1.isig[ind1 - 1 : ind1 + r + 1].data,
                    kind=kind,
                )
                f2 = interp1d(
                    axis2.axis[ind2 - r - 1 : ind2 + 1],
                    S2.isig[ind2 - r - 1 : ind2 + 1].data,
                    kind=kind,
                )
                length1 = axis.axis[ind1 - r : ind1 + 1].size
                grad1 = 0.5 / (length1 - 1)
                vect1 = np.arange(length1)
                length2 = axis2.axis[ind2 : ind2 + r].size
                if length2 == 1:
                    grad2 = 0.5
                else:
                    grad2 = 0.5 + 0.5 / (length2 - 1)
                vect2 = np.arange(length2)
                S1.data = np.hstack(
                    (
                        S1.isig[: ind1 - r].data,
                        (1 - grad1 * vect1) * S1.isig[ind1 - r : ind1 + 1].data
                        + grad1 * vect1 * f2(axis.axis[ind1 - r : ind1 + 1]),
                        (1 - grad2 * vect2) * f1(axis2.axis[ind2 : ind2 + r])
                        + grad2 * vect2 * S2.isig[ind2 : ind2 + r].data,
                        S2.isig[ind2 + r :].data,
                    )
                )
            else:  # just join at center of overlap
                S1.data = np.hstack((S1.isig[: ind1 + 1].data, S2.isig[ind2:].data))
    return S1


GRATING_EQUATION_DOCSTRING_PARAMETERS = r"""
    gamma_deg: float
        Inclination angle between the focal plane and the centre of the grating
        (found experimentally from calibration). In degree.
    deviation_angle_deg: float
        Also known as included angle. It is defined as the difference between
        angle of diffraction (:math:`\beta`) and angle of incidence
        (:math:`\alpha`). Given by manufacturer specsheet. In degree.
    focal_length_mm: float
        Given by manufacturer specsheet. In mm.
    ccd_width_mm: float
        The width of the CDD. Given by manufacturer specsheet. In mm.
    grating_central_wavelength_nm: float
        Wavelength at the centre of the grating, where exit slit is placed. In nm.
    grating_density_gr_mm: int
        Grating density in gratings per mm."""


def solve_grating_equation(
    axis,
    gamma_deg,
    deviation_angle_deg,
    focal_length_mm,
    ccd_width_mm,
    grating_central_wavelength_nm,
    grating_density_gr_mm,
):
    r"""Solves the grating equation.
    See `horiba.com/uk/scientific/products/optics-tutorial/wavelength-pixel-position`
    for equations.

    Parameters
    ----------
    axis: hyperspy.axis
        Axis in pixel units (no units) to convert to wavelength.
    %s

    Returns
    -------
    axis: hyperspy.axis
        HyperSpy axis object.
    """

    # From axis --> x-array
    s = str(axis.units).lower()
    non_defined = s in "<undefined>"
    pixel_units = s in ["pixel", "px"]

    if not non_defined and not pixel_units:
        warn(
            "The signal axes are already in {} units (not in pixel units)."
            "The conversion will run anyways.".format(str(axis.units)),
            SyntaxWarning,
        )

    # Set up variables
    # ch: channels,
    ch = len(axis.axis)

    # Calculate geometry
    # l_h: Perpendicular distance from the spectral plane to grating (Eq. 5.3)
    l_h = focal_length_mm * np.cos(np.deg2rad(gamma_deg))
    # h_blc: Distance from the intercept of the normal to the focal plane to the wavelength lambda_c (Eq. 5.5)
    h_blc = focal_length_mm * np.sin(np.deg2rad(gamma_deg))

    # alpha = angle of incidence (Eq. 2.1)
    numerator = 1e-6 * grating_density_gr_mm * grating_central_wavelength_nm
    denominator = 2 * np.cos(np.deg2rad(deviation_angle_deg / 2))
    alpha = np.arcsin(np.deg2rad(numerator / denominator)) - deviation_angle_deg / 2

    # beta: angle of diffraction (Eq. 1.2)
    beta = alpha + deviation_angle_deg

    # Find beta at the ends of the grating (Eq. 5.7)
    beta_min = (
        beta
        + np.deg2rad(gamma_deg)
        - np.arctan((ccd_width_mm / ch * 0.5 * ch - h_blc) / l_h)
    )
    beta_max = (
        beta
        + np.deg2rad(gamma_deg)
        - np.arctan((ccd_width_mm / ch * (1 - ch / 2) - h_blc) / l_h)
    )

    # Find lambda max/min given beta (Eq. 5.2)
    l_min = 1e6 * (np.sin(alpha) + np.sin(beta_min)) / grating_density_gr_mm
    l_max = 1e6 * (np.sin(alpha) + np.sin(beta_max)) / grating_density_gr_mm
    l_min = abs(l_min)
    l_max = abs(l_max)

    # Correct for diffraction index of air (lambda not in vacuum)
    l_min *= 1 / _n_air(l_min)
    l_max *= 1 / _n_air(l_max)

    # Create axis object to return
    scale = (l_max - l_min) / ch

    axis_class = UniformDataAxis

    axis_nm = axis_class(
        scale=scale,
        offset=l_min,
        name="Wavelength",
        units="nm",
        navigate=False,
        size=ch,
    )
    return axis_nm


solve_grating_equation.__doc__ %= GRATING_EQUATION_DOCSTRING_PARAMETERS



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from numpy import ones, arange
from numpy.random import random
from numpy.testing import assert_allclose
from pytest import raises, mark, skip, warns

from hyperspy.axes import DataAxis
from lumispy import join_spectra, nm2eV, eV2nm, nm2invcm, invcm2nm
from lumispy.utils.axes import _n_air
from lumispy.signals import LumiSpectrum


def test__n_air():
    wl = arange(800) * 2 + 150
    with warns(UserWarning, match="The wavelength"):
        n = _n_air(wl)
    assert n[0] == _n_air(185)
    assert n[-1] == _n_air(1700)
    with warns(UserWarning):
        assert _n_air(180) == _n_air(185)
    with warns(UserWarning):
        assert _n_air(1705) == _n_air(1700)
    assert_allclose(_n_air(200), 1.00032406)
    assert_allclose(_n_air(500), 1.00027896)


def test_nm2eV():
    assert_allclose(nm2eV(200), 6.19720164)
    assert_allclose(nm2eV(300), 4.13160202)


def test_eV2nm():
    assert_allclose(eV2nm(2), 619.74951462)
    assert_allclose(eV2nm(3), 413.16411768)


def test_nm2invcm():
    assert nm2invcm(250) == 4e4
    assert nm2invcm(400) == 25e3


def test_invcm2nm():
    assert invcm2nm(4e4) == 250
    assert invcm2nm(25e3) == 400


@mark.parametrize(("average"), (True, False))
@mark.parametrize(("scale"), (True, False))
@mark.parametrize(("kind"), ("slinear", "linear"))
def test_joinspectra(average, scale, kind):
    s1 = LumiSpectrum(arange(32))
    s2 = LumiSpectrum(arange(32) + 25)
    s3 = LumiSpectrum(arange(32) + 50)
    s2.axes_manager.signal_axes[0].offset = 25
    s3.axes_manager.signal_axes[0].offset = 50
    s = join_spectra([s1, s2, s3], r=2, average=average, scale=scale, kind=kind)
    assert s.data[-1] == 81
    assert s.axes_manager.signal_axes[0].scale == 1
    assert s.axes_manager.signal_axes[0].size == 82


def test_joinspectra2():
    s1 = LumiSpectrum(arange(32))
    s2 = LumiSpectrum(arange(32) + 25)
    s2.axes_manager.signal_axes[0].offset = 25
    s2.isig[3] = 0
    s = join_spectra([s1, s2], r=2, average=True, scale=True)
    assert s.data[-1] == 56
    assert s.data[28] == 28 / 3


def test_joinspectra_length1():
    s1 = LumiSpectrum(arange(32))
    s2 = LumiSpectrum(arange(32) + 27)
    s2.axes_manager.signal_axes[0].offset = 27
    s = join_spectra([s1, s2], r=1, average=False, scale=True)
    assert s.data[-1] == 58
    with raises(ValueError, match="Averaging can not be performed for r=1."):
        s = join_spectra([s1, s2], r=1, average=True, scale=True)
    s1.axes_manager[0].convert_to_non_uniform_axis()
    s2.axes_manager[0].convert_to_non_uniform_axis()
    s = join_spectra([s1, s2], r=1, average=True, scale=True)
    assert s.data[-1] == 58


def test_joinspectra_errors():
    s1 = LumiSpectrum(ones(32))
    s2 = LumiSpectrum(ones(32) * 2)
    s2.axes_manager.signal_axes[0].offset = 25
    # Test that catch for r works
    raises(ValueError, join_spectra, [s1, s2])
    s2.axes_manager.signal_axes[0].offset = 35
    # Test that overlap catch works
    raises(ValueError, join_spectra, [s1, s2], r=2)
    s1.data *= -1
    s2.axes_manager.signal_axes[0].offset = 25
    raises(ValueError, join_spectra, [s1, s2], r=2)


@mark.parametrize(("average"), (True, False))
@mark.parametrize(("scale"), (True, False))
@mark.parametrize(("kind"), ("slinear", "linear"))
def test_joinspectra_linescan(average, scale, kind):
    s1 = LumiSpectrum(random((4, 64)))
    s2 = LumiSpectrum(random((4, 64)))
    s2.axes_manager.signal_axes[0].offset = 47
    s = join_spectra([s1, s2], r=7, average=average, scale=scale, kind=kind)
    assert s.axes_manager.signal_axes[0].size == 111
    assert s.axes_manager.signal_axes[0].scale == 1


@mark.parametrize(("average"), (True, False))
@mark.parametrize(("scale"), (True, False))
@mark.parametrize(("kind"), ("slinear", "linear"))
def test_joinspectra_nonuniform(average, scale, kind):
    s1 = LumiSpectrum(arange(32))
    s2 = LumiSpectrum(arange(32) + 25)
    s2.axes_manager.signal_axes[0].offset = 25
    s1.axes_manager.signal_axes[0].convert_to_non_uniform_axis()
    s = join_spectra([s1, s2], r=2, average=average, scale=scale, kind=kind)
    assert s.axes_manager.signal_axes[0].is_uniform == False
    assert s.axes_manager.signal_axes[0].size == 57
    assert s.axes_manager.signal_axes[0].axis[-1] == 56
    assert s.data.size == 57
    assert s.data[-1] == 56
    s1 = LumiSpectrum(arange(12))
    s2 = LumiSpectrum(arange(12) + 3.8, axes=[DataAxis(axis=arange(12) + 3.8)])
    s = join_spectra([s1, s2], r=2, average=average, scale=scale, kind=kind)
    assert s.axes_manager[0].axis.size == 16
    assert s.data.size == 16
    assert s.data[-1] == 14.8


@mark.parametrize(("average"), (True, False))
@mark.parametrize(("scale"), (True, False))
@mark.parametrize(("kind"), ("slinear", "linear"))
def test_joinspectra_FunctionalDA(average, scale, kind):
    try:
        from hyperspy.axes import FunctionalDataAxis
    except ImportError:
        skip("HyperSpy version doesn't support non-uniform axis")
    s1 = LumiSpectrum(ones(32))
    s2 = LumiSpectrum(ones(32) * 2)
    s2.axes_manager.signal_axes[0].offset = 25
    s1.axes_manager.signal_axes[0].convert_to_functional_data_axis(expression="x**2")
    s2.axes_manager.signal_axes[0].convert_to_functional_data_axis(expression="x**2")
    s = join_spectra([s1, s2], r=2, average=average, scale=scale, kind=kind)
    assert s.axes_manager.signal_axes[0].is_uniform == False
    assert s.axes_manager.signal_axes[0].size == 57
    assert s.axes_manager.signal_axes[0].axis[-1] == 3136
    assert s.data.size == 57
    if scale:
        assert s.data[-1] == 1
    else:
        assert s.data[-1] == 2
    # test that join_spectra works for r that is float not int
    s = join_spectra([s1, s2], r=2.1, average=average, scale=scale, kind=kind)



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
Signal class for electroluminescence spectral data
--------------------------------------------------
"""

from hyperspy._signals.lazy import LazySignal

from lumispy.signals import LumiSpectrum


class ELSpectrum(LumiSpectrum):
    """**General 1D electroluminescence signal class**"""

    _signal_type = "EL"
    _signal_dimension = 1


class LazyELSpectrum(LazySignal, ELSpectrum):
    """**General lazy 1D electroluminescence signal class**"""

    _lazy = True



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest
from numpy import arange

from hyperspy.signals import Signal1D, Signal2D

from lumispy.signals import (
    LumiSpectrum,
    CLSpectrum,
    CLSEMSpectrum,
    CLSTEMSpectrum,
    PLSpectrum,
    ELSpectrum,
    LumiTransient,
    LumiTransientSpectrum,
    LazyLumiSpectrum,
    LazyCLSpectrum,
    LazyCLSEMSpectrum,
    LazyCLSTEMSpectrum,
    LazyPLSpectrum,
    LazyELSpectrum,
    LazyLumiTransient,
    LazyLumiTransientSpectrum,
)

signal1d_class_list = [
    LumiSpectrum,
    CLSpectrum,
    CLSEMSpectrum,
    CLSTEMSpectrum,
    PLSpectrum,
    ELSpectrum,
    LazyLumiSpectrum,
    LazyCLSpectrum,
    LazyCLSEMSpectrum,
    LazyCLSTEMSpectrum,
    LazyPLSpectrum,
    LazyELSpectrum,
    LumiTransient,
    LazyLumiTransient,
]
signal2d_class_list = [
    LumiTransientSpectrum,
    LazyLumiTransientSpectrum,
]


@pytest.mark.parametrize("signal_class", signal1d_class_list)
def test_signal_registration1d(signal_class):
    s = signal_class(0)

    s2 = Signal1D([0, 1, 2])
    if s._lazy:
        s2 = s2.as_lazy()

    print("signal_type:", s._signal_type)

    s2.set_signal_type(s._signal_type)
    assert isinstance(s2, signal_class)


@pytest.mark.parametrize("signal_class", signal2d_class_list)
def test_signal_registration2d(signal_class):
    data = arange(1000).reshape((10, 10, 10))
    s = signal_class(data)

    s2 = Signal2D(data)
    if s._lazy:
        s2 = s2.as_lazy()

    print("signal_type:", s._signal_type)

    s2.set_signal_type(s._signal_type)
    assert isinstance(s2, signal_class)



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
Signal class for luminescence transient data (1D)
-------------------------------------------------
"""

from hyperspy.signals import Signal1D
from hyperspy._signals.lazy import LazySignal

from lumispy.signals.common_transient import CommonTransient


class LumiTransient(Signal1D, CommonTransient):
    """**General 1D luminescence signal class (transient/time resolved)**"""

    _signal_type = "Transient"
    _signal_dimension = 1

    def __init(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LazyLumiTransient(LazySignal, LumiTransient):
    """**General lazy 1D luminescence signal class (transient/time resolved)**"""

    _lazy = True



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

from lumispy.signals.luminescence_spectrum import LumiSpectrum
from hyperspy._signals.signal2d import Signal2D
from numpy.testing import assert_allclose

backgrounds = [
    ([np.ones(50)], [np.zeros(50, dtype="float64")]),
    (
        [np.linspace(0, 49, num=50, dtype="float64"), np.ones(50)],
        [np.zeros(50, dtype="float64")],
    ),
    (
        [np.linspace(0, 50, num=30, dtype="float64"), np.ones(30)],
        [np.zeros(50, dtype="float64")],
    ),
    (LumiSpectrum(np.ones(50)), [np.zeros(50, dtype="float64")]),
]


class TestLumiSpectrum:
    def test_remove_background_from_file(self):
        for bkg, output in backgrounds:
            s = LumiSpectrum(np.ones(50))
            s2 = s.remove_background_from_file(bkg, inplace=False)
            s.remove_background_from_file(bkg, inplace=True)
            assert np.allclose(s.data, output)
            assert np.allclose(s2.data, output)
            assert s.metadata.Signal.background_subtracted is True
            assert hasattr(s.metadata.Signal, "background")

    def test_errors_raise(self):
        s = LumiSpectrum(np.ones(50))
        with pytest.raises(AttributeError):
            bkg = np.array([[1, 1], [1, 1], [1, 1]])
            s.remove_background_from_file(bkg)
        # Test that a GUI is opened if s.remove_background_from_file is passed without a background
        # s.remove_background_from_file()
        # Test double background removal
        s.remove_background_from_file(backgrounds[0][0], inplace=True)
        with pytest.raises(RecursionError):
            s.remove_background_from_file(backgrounds[0][0])

    def test_warnings(self):
        pytest.importorskip("hyperspy_gui_ipywidgets")
        s = LumiSpectrum(np.ones(50))
        with pytest.warns(SyntaxWarning, match="Using the Hyperspy"):
            s.remove_background_from_file(background=None, display=False)

    def test_deprecation_warning(self):
        s = LumiSpectrum(np.ones(50))
        with pytest.warns(DeprecationWarning, match="deprecated"):
            s.remove_background_from_file(background=backgrounds[0][0])

    def test_px_to_nm_grating_solver(self):
        s = LumiSpectrum(np.ones(10))
        ax = s.axes_manager.signal_axes[0]
        ax.offset = 200
        ax.scale = 10

        s_copy = s.px_to_nm_grating_solver(
            3,
            -20,
            300,
            25,
            600,
            150,
        )
        s.px_to_nm_grating_solver(3, -20, 300, 25, 600, 150, inplace=True)

        assert s_copy.axes_manager.signal_axes[0].name == "Wavelength"
        assert s_copy.axes_manager.signal_axes[0].units == "nm"
        assert s.axes_manager.signal_axes[0].name == "Wavelength"
        assert s.axes_manager.signal_axes[0].units == "nm"

        assert_allclose(
            s_copy.axes_manager.signal_axes[0].axis[0],
            368.614,
            atol=0.1,
        )
        assert_allclose(
            s_copy.axes_manager.signal_axes[0].axis[-1],
            768.249,
            atol=0.1,
        )
        assert_allclose(s.axes_manager.signal_axes[0].axis[0], 368.614, atol=0.1)
        assert_allclose(s.axes_manager.signal_axes[0].axis[-1], 768.249, atol=0.1)

    def test_center_of_mass(self):
        s = LumiSpectrum([1, 2, 3, 2, 1, 0])
        ax = s.axes_manager.signal_axes[0]
        ax.offset = 200
        ax.scale = 100
        ax.units = "nm"
        ax.name = "Wavelength"
        s.metadata.General.title = "test_signal"

        com = s.centroid()
        assert_allclose(com.data, 400.0, atol=0.1)
        assert (
            com.metadata.General.title
            == f"Centroid map of {ax.name} ({ax.units}) for test_signal"
        )

    def test_center_of_mass_signalrange(self):
        s = LumiSpectrum([100, 100, 1, 2, 3, 2, 1, 0, 100, 100])
        ax = s.axes_manager.signal_axes[0]
        ax.offset = 0
        ax.scale = 100
        ax.units = "nm"
        ax.name = "Wavelength"

        com = s.centroid(signal_range=(2, -2))
        assert_allclose(com.data, 400.0, atol=0.1)
        com = s.centroid(signal_range=(200.0, 800.0))
        assert_allclose(com.data, 400.0, atol=0.1)
        assert com.metadata.General.title == f"Centroid map of {ax.name} ({ax.units})"
        with pytest.raises(TypeError):
            s.centroid(signal_range=(1))
        with pytest.raises(TypeError):
            s.centroid(signal_range="string")
        with pytest.raises(ValueError):
            s.centroid(signal_range=(1, 2, 3))

    def test_centre_of_mass_3d(self):
        s = LumiSpectrum([[[1, 2, 3, 4, 5]] * 3] * 4)
        com = s.centroid()
        assert com.axes_manager.shape == (
            3,
            4,
        )
        assert com.metadata.General.title == f"Centroid map"
        assert type(com) == Signal2D



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import hyperspy
import numpy as np
import pytest

from lumispy.signals import CLSEMSpectrum


class TestCLSEMSpectrum:
    @pytest.mark.skipif(
        hyperspy.__version__ == "1.6.2", reason="Broken with hyperspy 1.6.2"
    )
    @pytest.mark.parametrize("nx, ny", [(10, 20), (20, 10)])
    def test_correct_grating_shift(self, nx, ny):
        calx, corg, fov = 1e-10, 1e-10, 1e-10
        s = CLSEMSpectrum(np.random.random(nx * ny * 100).reshape(ny, nx, 100))

        garray = np.arange(
            (-corg / 2) * calx / (fov * nx) * 1000 * nx,
            (corg / 2) * calx / (fov * nx) * 1000 * nx,
            corg * calx / (fov * nx) * 1000,
        )
        barray = np.full((ny, nx), garray)

        s2 = s.deepcopy()
        s.correct_grating_shift(calx, corg, fov)
        s2.shift1D(barray)
        np.testing.assert_allclose(s2.data, s.data)

    @pytest.mark.skipif(
        hyperspy.__version__ == "1.6.2", reason="Broken with hyperspy 1.6.2"
    )
    def test_double_correct_grating_shift(self):
        s = CLSEMSpectrum(np.ones((10, 10, 10)))
        s.correct_grating_shift(1e-10, 1e-10, 1e-10)
        with pytest.raises(BaseException):
            s.correct_grating_shift(1, 1, 1)



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

from inspect import getfullargspec

from lumispy.signals.cl_spectrum import CLSpectrum


param_list_signal_mask = [
    ([900, 500], [False, False, False, True, True, True, True, True, True, False]),
    (
        [[500, 100], [700, 1]],
        [False, True, True, False, True, False, False, False, False, False],
    ),
    (
        [[200, 100], [600, 300]],
        [True, True, True, True, True, False, False, False, False, False],
    ),
]


class TestCLSpectrum:
    def test__make_signal_mask(self):
        s = CLSpectrum(np.ones(10))
        s.axes_manager.signal_axes[0].scale = 100.5
        s.axes_manager.signal_axes[0].offset = 300
        for peak_list, mask_test in param_list_signal_mask:
            mask = s._make_signal_mask(peak_list)
            np.testing.assert_allclose(mask, mask_test)

    def test_remove_spikes(self):
        s = CLSpectrum(np.ones((2, 3, 30)))
        s.add_gaussian_noise(1e-5)
        # Add three spikes
        s.data[1, 0, 1] += 2
        s.data[0, 2, 29] += 1.001

        if not "threshold" in getfullargspec(s.spikes_removal_tool)[0]:
            try:
                s.remove_spikes()
            except ImportError:
                pytest.skip("HyperSpy version doesn't support spike removal.")

        with pytest.raises(AttributeError):
            s.remove_spikes(luminescence_roi=[1], signal_mask=[1])

        with pytest.warns(UserWarning, match="Threshold value: 1.00"):
            s1 = s.remove_spikes()

        np.testing.assert_almost_equal(s1.data[1, 0, 1], 1, decimal=4)
        np.testing.assert_almost_equal(s1.data[0, 2, 29], 1, decimal=4)

        s3 = s.remove_spikes(show_diagnosis_histogram=True)
        hist_data = s._spikes_diagnosis(
            max_num_bins=25,
            show_plot=False,
            use_gui=False,
        )
        expected_data = np.zeros(25)
        expected_data[0] = 176
        expected_data[6] = 1
        expected_data[12] = 2
        expected_data[-1] = 1
        np.testing.assert_allclose(hist_data.data, expected_data)
        np.testing.assert_almost_equal(s3.data[1, 0, 1], 1, decimal=4)
        np.testing.assert_almost_equal(s3.data[0, 2, 29], 1, decimal=4)

        lum_roi = [1, 1]
        s4 = s.remove_spikes(luminescence_roi=lum_roi, threshold=0.5)
        np.testing.assert_almost_equal(s4.data[1, 0, 1], 3, decimal=4)
        np.testing.assert_almost_equal(s4.data[0, 2, 29], 1, decimal=4)

        s.remove_spikes(inplace=True, threshold=0.5)
        np.testing.assert_almost_equal(s.data[1, 0, 1], 1, decimal=4)
        np.testing.assert_almost_equal(s.data[0, 2, 29], 1, decimal=4)
        # TODO: test if histogram is shown as a plot if show_diagnosis_histogram=True.



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from numpy import arange, ones
from numpy.testing import assert_allclose
from pytest import raises, mark, warns

from hyperspy.axes import DataAxis, UniformDataAxis
from lumispy.signals import LumiSpectrum, LumiTransientSpectrum
from lumispy.utils.axes import (
    nm2eV,
    eV2nm,
    axis2eV,
    data2eV,
    var2eV,
    nm2invcm,
    invcm2nm,
    axis2invcm,
    data2invcm,
    var2invcm,
    solve_grating_equation,
)


def test_nm2eV():
    wl = arange(300, 400, 90)
    en = nm2eV(wl)
    assert_allclose(en[0], 4.13160202)
    assert_allclose(en[-1], 3.17818160)


def test_eV2nm():
    en = arange(1, 2, 0.8)
    wl = eV2nm(en)
    assert_allclose(wl[0], 1239.50284)
    assert_allclose(wl[-1], 688.611116)


def test_axis2eV():
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    axis2 = DataAxis(axis=arange(0.2, 0.400, 0.01), units="µm")
    axis3 = DataAxis(axis=arange(1, 2, 0.1), units="eV")
    evaxis, factor = axis2eV(axis)
    evaxis2, factor2 = axis2eV(axis2)
    with raises(AttributeError, match="Signal unit is already eV."):
        axis2eV(axis3)
    assert factor == 1e6
    assert factor2 == 1e3
    assert evaxis.name == "Energy"
    assert evaxis.units == "eV"
    assert not evaxis.navigate
    assert evaxis2.units == "eV"
    assert evaxis2.size == 20
    assert_allclose(evaxis.axis[0], evaxis2.axis[0])
    assert_allclose(evaxis.axis[-1], evaxis2.axis[-1])
    assert_allclose(evaxis.axis[0], 3.1781816)


def test_data2eV():
    data = 100 * ones(20)
    ax0 = DataAxis(axis=arange(200, 400, 10), units="nm")
    evaxis, factor = axis2eV(ax0)
    evdata = data2eV(data, factor, evaxis.axis, ax0)
    assert_allclose(evdata[0], 12.271168)
    ax0 = DataAxis(axis=arange(0.2, 0.4, 0.01), units="µm")
    evaxis, factor = axis2eV(ax0)
    evdata = data2eV(data, factor, evaxis.axis, ax0)
    assert_allclose(evdata[0], 12.271168e-3)


def test_var2eV():
    data = 100 * ones(20)
    ax0 = DataAxis(axis=arange(200, 400, 10), units="nm")
    evaxis, factor = axis2eV(ax0)
    evvar = var2eV(data, factor, evaxis.axis, ax0)
    assert_allclose(evvar[0], 1.5058156)


@mark.parametrize(("jacobian"), (True, False))
@mark.parametrize(("variance"), (True, False, "constant"))
def test_to_eV(jacobian, variance):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    data[0] += 1
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    if variance:
        if variance == "constant":
            S1.set_noise_variance(1.0)
        else:
            S1.estimate_poissonian_noise_variance()
    S2 = S1.to_eV(inplace=False, jacobian=jacobian)
    S1.axes_manager[0].units = "µm"
    S1.axes_manager[0].axis = axis.axis / 1000
    if jacobian:
        S1.data *= 1000
    S1.to_eV(jacobian=jacobian)
    assert S1.axes_manager[0].units == "eV"
    assert S2.axes_manager[0].name == "Energy"
    assert S2.axes_manager[0].size == 20
    assert S1.axes_manager[0].axis[0] == S2.axes_manager[0].axis[0]
    assert_allclose(S1.data, S2.data, 5e-4)
    if not jacobian:
        S1.data *= 10  # check that deepcopy worked
        assert S2.data[-1] == 2
        assert S1.data[-1] == 20
    else:
        S1.data *= 10
        assert_allclose(S2.data[-1], 0.06454526)
    nav = UniformDataAxis(size=4)
    # navigation dimension 1
    L1 = LumiSpectrum(
        ones((4, 20)), axes=[nav.get_axis_dictionary(), axis.get_axis_dictionary()]
    )
    if variance:
        if variance == "constant":
            L1.set_noise_variance(1.0)
        else:
            L1.estimate_poissonian_noise_variance()
    L2 = L1.to_eV(inplace=False, jacobian=jacobian)
    L1.to_eV(jacobian=jacobian)
    assert L1.axes_manager.signal_axes[0].units == "eV"
    assert L2.axes_manager.signal_axes[0].name == "Energy"
    assert L2.axes_manager.signal_axes[0].size == 20
    assert (
        L1.axes_manager.signal_axes[0].axis[0] == L2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(L1.data, L2.data, 5e-4)
    # navigation dimension 2
    M1 = LumiSpectrum(
        ones((4, 4, 20)),
        axes=[
            nav.get_axis_dictionary(),
            nav.get_axis_dictionary(),
            axis.get_axis_dictionary(),
        ],
    )
    if variance:
        if variance == "constant":
            M1.set_noise_variance(1.0)
        else:
            M1.estimate_poissonian_noise_variance()
    M2 = M1.to_eV(inplace=False, jacobian=jacobian)
    M1.to_eV(jacobian=jacobian)
    assert M1.axes_manager.signal_axes[0].units == "eV"
    assert M2.axes_manager.signal_axes[0].name == "Energy"
    assert M2.axes_manager.signal_axes[0].size == 20
    assert (
        M1.axes_manager.signal_axes[0].axis[0] == M2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(M1.data, M2.data, 5e-4)
    if variance:
        if variance != "constant":
            assert (
                S1.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S1.axes_manager[-1].axis[0]
            )
            assert (
                S2.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S2.axes_manager[-1].axis[0]
            )
        assert (
            S1.metadata.Signal.Noise_properties.variance
            == S2.metadata.Signal.Noise_properties.variance
        )
        assert (
            L1.metadata.Signal.Noise_properties.variance
            == L2.metadata.Signal.Noise_properties.variance
        )
        assert (
            M1.metadata.Signal.Noise_properties.variance
            == M2.metadata.Signal.Noise_properties.variance
        )
    else:
        assert S1.metadata.has_item("Signal.Noise_properties.variance") == False


def test_eV_slicing():
    S = LumiSpectrum(arange(100), axes=[{"axis": arange(100) + 300}])
    S.to_eV(inplace=True)
    S.isig[3.251:4.052]


def test_to_eV_2D():
    axis1 = UniformDataAxis(size=10, offset=200, scale=10, units="nm")
    axis2 = UniformDataAxis(size=8, offset=0, scale=0.1, units="ns")
    data = arange(80).reshape(8, 10)
    S1 = LumiTransientSpectrum(data, axes=[axis2, axis1])
    S2 = S1.to_eV(inplace=True, jacobian=False)
    assert S1.axes_manager[0].units == "eV"
    assert S1.axes_manager[0].axis[-1] == nm2eV(axis1.axis[0])
    assert S1.data[0, 0] == data[0, -1]


@mark.parametrize(("jacobian"), (True, False))
def test_reset_variance_linear_model_eV(jacobian):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    S1.metadata.set_item("Signal.Noise_properties.Variance_linear_model.gain_factor", 2)
    S1.metadata.set_item("Signal.Noise_properties.Variance_linear_model.gain_offset", 1)
    S1.metadata.set_item(
        "Signal.Noise_properties.Variance_linear_model.correlation_factor", 2
    )
    S1.estimate_poissonian_noise_variance()
    S2 = S1.to_eV(inplace=False, jacobian=jacobian)
    if jacobian:
        with warns(UserWarning, match="Following"):
            S1.to_eV(inplace=True, jacobian=jacobian)
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 1
        )
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_offset == 0
        )
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.correlation_factor
            == 1
        )
        assert (
            S2.metadata.has_item("Signal.Noise_properties.Variance_linear_model")
            == False
        )
    else:
        S1.to_invcm(inplace=True, jacobian=jacobian)
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 2
        )
        assert (
            S2.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 2
        )


def test_nm2invcm():
    wl = arange(300, 410, 100)
    invcm = nm2invcm(wl)
    assert_allclose(invcm[0], 33333.3333)
    assert_allclose(invcm[-1], 25000)


def test_invcm2nm():
    invcm = arange(10000, 20000, 6000)
    wl = invcm2nm(invcm)
    assert_allclose(wl[0], 1000)
    assert_allclose(wl[-1], 625)


def test_axis2invcm():
    axis = UniformDataAxis(size=21, offset=200, scale=10)
    axis2 = DataAxis(axis=arange(0.2, 0.410, 0.01), units="µm")
    axis3 = DataAxis(axis=arange(1, 2, 0.1), units=r"cm$^{-1}$")
    invcmaxis, factor = axis2invcm(axis)
    invcmaxis2, factor2 = axis2invcm(axis2)
    with raises(AttributeError, match="Signal unit is already"):
        axis2invcm(axis3)
    assert factor == 1e7
    assert factor2 == 1e4
    assert invcmaxis.name == "Wavenumber"
    assert invcmaxis.units == r"cm$^{-1}$"
    assert not invcmaxis.navigate
    assert invcmaxis2.units == r"cm$^{-1}$"
    assert invcmaxis2.size == 21
    assert_allclose(invcmaxis.axis[0], invcmaxis2.axis[0])
    assert_allclose(invcmaxis.axis[-1], invcmaxis2.axis[-1])
    assert_allclose(invcmaxis.axis[0], 25000)


def test_data2invcm():
    data = 100 * ones(20)
    factor = 1e7
    ax0 = arange(200, 400, 10)
    invcmaxis = nm2invcm(ax0)
    invcmdata = data2invcm(data, factor, invcmaxis)
    assert_allclose(invcmdata[-1], 1.521)


def test_var2invcm():
    data = 100 * ones(20)
    factor = 1e7
    ax0 = arange(200, 400, 10)
    invcmaxis = nm2invcm(ax0)
    invcmdata = var2invcm(data, factor, invcmaxis)
    assert_allclose(invcmdata[-1], 0.02313441)


@mark.parametrize(("jacobian"), (True, False))
@mark.parametrize(("variance"), (True, False, "constant"))
def test_to_invcm(jacobian, variance):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    data[0] += 1
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    if variance:
        if variance == "constant":
            S1.set_noise_variance(1.0)
        else:
            S1.estimate_poissonian_noise_variance()
    S2 = S1.to_invcm(inplace=False, jacobian=jacobian)
    S1.axes_manager[0].units = "µm"
    S1.axes_manager[0].axis = axis.axis / 1000
    if jacobian:
        S1.data *= 1000
    S1.to_invcm(jacobian=jacobian)
    assert S1.axes_manager[0].units == r"cm$^{-1}$"
    assert S2.axes_manager[0].name == "Wavenumber"
    assert S2.axes_manager[0].size == 20
    assert S1.axes_manager[0].axis[0] == S2.axes_manager[0].axis[0]
    assert_allclose(S1.data, S2.data, 5e-4)
    if not jacobian:
        S1.data *= 10  # check that deepcopy worked
        assert S2.data[-1] == 2
        assert S1.data[-1] == 20
    else:
        S1.data *= 10
        assert S2.data[-1] == 0.008
    nav = UniformDataAxis(size=4)
    # navigation dimension 1
    L1 = LumiSpectrum(
        ones((4, 20)), axes=[nav.get_axis_dictionary(), axis.get_axis_dictionary()]
    )
    if variance:
        if variance == "constant":
            L1.set_noise_variance(1.0)
        else:
            L1.estimate_poissonian_noise_variance()
    L2 = L1.to_invcm(inplace=False, jacobian=jacobian)
    L1.to_invcm(jacobian=jacobian)
    assert L1.axes_manager.signal_axes[0].units == r"cm$^{-1}$"
    assert L2.axes_manager.signal_axes[0].name == "Wavenumber"
    assert L2.axes_manager.signal_axes[0].size == 20
    assert (
        L1.axes_manager.signal_axes[0].axis[0] == L2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(L1.data, L2.data, 5e-4)
    # navigation dimension 2
    M1 = LumiSpectrum(
        ones((4, 4, 20)),
        axes=[
            nav.get_axis_dictionary(),
            nav.get_axis_dictionary(),
            axis.get_axis_dictionary(),
        ],
    )
    if variance:
        if variance == "constant":
            M1.set_noise_variance(1.0)
        else:
            M1.estimate_poissonian_noise_variance()
    M2 = M1.to_invcm(inplace=False, jacobian=jacobian)
    M1.to_invcm(jacobian=jacobian)
    assert M1.axes_manager.signal_axes[0].units == r"cm$^{-1}$"
    assert M2.axes_manager.signal_axes[0].name == "Wavenumber"
    assert M2.axes_manager.signal_axes[0].size == 20
    assert (
        M1.axes_manager.signal_axes[0].axis[0] == M2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(M1.data, M2.data, 5e-4)
    if variance:
        if variance != "constant":
            assert (
                S1.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S1.axes_manager[-1].axis[0]
            )
            assert (
                S2.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S2.axes_manager[-1].axis[0]
            )
        assert (
            S1.metadata.Signal.Noise_properties.variance
            == S2.metadata.Signal.Noise_properties.variance
        )
        assert (
            L1.metadata.Signal.Noise_properties.variance
            == L2.metadata.Signal.Noise_properties.variance
        )
        assert (
            M1.metadata.Signal.Noise_properties.variance
            == M2.metadata.Signal.Noise_properties.variance
        )
    else:
        assert S1.metadata.has_item("Signal.Noise_properties.variance") == False


def test_invcm_slicing():
    S = LumiSpectrum(arange(100), axes=[{"axis": arange(100) + 300}])
    S.to_invcm(inplace=True)
    S.isig[26000.0:30000.0]


@mark.parametrize(("jacobian"), (True, False))
def test_reset_variance_linear_model_invcm(jacobian):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    S1.metadata.set_item("Signal.Noise_properties.Variance_linear_model.gain_factor", 2)
    S1.metadata.set_item("Signal.Noise_properties.Variance_linear_model.gain_offset", 1)
    S1.metadata.set_item(
        "Signal.Noise_properties.Variance_linear_model.correlation_factor", 2
    )
    S1.estimate_poissonian_noise_variance()
    S2 = S1.to_invcm(inplace=False, jacobian=jacobian)
    if jacobian:
        with warns(UserWarning, match="Following"):
            S1.to_invcm(inplace=True, jacobian=jacobian)
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 1
        )
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_offset == 0
        )
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.correlation_factor
            == 1
        )
        assert (
            S2.metadata.has_item("Signal.Noise_properties.Variance_linear_model")
            == False
        )
    else:
        S1.to_invcm(inplace=True, jacobian=jacobian)
        assert (
            S1.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 2
        )
        assert (
            S2.metadata.Signal.Noise_properties.Variance_linear_model.gain_factor == 2
        )


@mark.parametrize(("jacobian"), (True, False))
@mark.parametrize(("variance"), (True, False, "constant"))
def test_to_invcm_relative(jacobian, variance):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    data[0] += 1
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    if variance:
        if variance == "constant":
            S1.set_noise_variance(1.0)
        else:
            S1.estimate_poissonian_noise_variance()
    S2 = S1.to_invcm_relative(laser=244, inplace=False, jacobian=jacobian)
    S1.axes_manager[0].units = "µm"
    S1.axes_manager[0].axis = axis.axis / 1000
    if jacobian:
        S1.data *= 1000
    S1.to_invcm_relative(laser=0.244, jacobian=jacobian)
    assert S1.axes_manager[0].units == r"cm$^{-1}$"
    assert S2.axes_manager[0].name == "Raman Shift"
    assert S2.axes_manager[0].size == 20
    assert S1.axes_manager[0].axis[0] == S2.axes_manager[0].axis[0]
    assert_allclose(S1.data, S2.data, 5e-4)
    nav = UniformDataAxis(size=4)
    # navigation dimension 1
    L1 = LumiSpectrum(
        ones((4, 20)), axes=[nav.get_axis_dictionary(), axis.get_axis_dictionary()]
    )
    if variance:
        if variance == "constant":
            L1.set_noise_variance(1.0)
        else:
            L1.estimate_poissonian_noise_variance()
    L2 = L1.to_invcm_relative(laser=244, inplace=False, jacobian=jacobian)
    L1.to_invcm_relative(laser=244, jacobian=jacobian)
    assert L1.axes_manager.signal_axes[0].units == r"cm$^{-1}$"
    assert L2.axes_manager.signal_axes[0].name == "Raman Shift"
    assert L2.axes_manager.signal_axes[0].size == 20
    assert (
        L1.axes_manager.signal_axes[0].axis[0] == L2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(L1.data, L2.data, 5e-4)
    # navigation dimension 2
    M1 = LumiSpectrum(
        ones((4, 4, 20)),
        axes=[
            nav.get_axis_dictionary(),
            nav.get_axis_dictionary(),
            axis.get_axis_dictionary(),
        ],
    )
    if variance:
        if variance == "constant":
            M1.set_noise_variance(1.0)
        else:
            M1.estimate_poissonian_noise_variance()
    M2 = M1.to_invcm_relative(laser=244, inplace=False, jacobian=jacobian)
    M1.to_invcm_relative(laser=244, jacobian=jacobian)
    assert M1.axes_manager.signal_axes[0].units == r"cm$^{-1}$"
    assert M2.axes_manager.signal_axes[0].name == "Raman Shift"
    assert M2.axes_manager.signal_axes[0].size == 20
    assert (
        M1.axes_manager.signal_axes[0].axis[0] == M2.axes_manager.signal_axes[0].axis[0]
    )
    assert_allclose(M1.data, M2.data, 5e-4)
    if variance:
        if variance != "constant":
            assert (
                S1.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S1.axes_manager[-1].axis[0]
            )
            assert (
                S2.metadata.Signal.Noise_properties.variance.axes_manager[-1].axis[0]
                == S2.axes_manager[-1].axis[0]
            )
        assert (
            S1.metadata.Signal.Noise_properties.variance
            == S2.metadata.Signal.Noise_properties.variance
        )
        assert (
            L1.metadata.Signal.Noise_properties.variance
            == L2.metadata.Signal.Noise_properties.variance
        )
        assert (
            M1.metadata.Signal.Noise_properties.variance
            == M2.metadata.Signal.Noise_properties.variance
        )
    else:
        assert S1.metadata.has_item("Signal.Noise_properties.variance") == False


@mark.parametrize(("jacobian"), (True, False))
def test_to_raman_shift(jacobian):
    axis = UniformDataAxis(size=20, offset=200, scale=10)
    data = ones(20)
    data[0] += 1
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    S2 = S1.to_raman_shift(laser=244, inplace=False, jacobian=jacobian)
    S1.axes_manager[0].units = "µm"
    S1.axes_manager[0].axis = axis.axis / 1000
    if jacobian:
        S1.data *= 1000
    S1.to_raman_shift(laser=0.244, jacobian=jacobian)
    assert S1.axes_manager[0].units == r"cm$^{-1}$"
    assert S2.axes_manager[0].name == "Raman Shift"
    assert S2.axes_manager[0].size == 20
    assert S1.axes_manager[0].axis[0] == S2.axes_manager[0].axis[0]
    assert_allclose(S1.data, S2.data, 5e-4)
    if not jacobian:
        S1.data *= 10  # check that deepcopy worked
        assert S2.data[0] == 2
        assert S1.data[0] == 20
    else:
        S1.data *= 10
        assert S2.data[0] == 0.008


def test_to_raman_shift_laser():
    axis = UniformDataAxis(size=20, offset=200, scale=10, units="nm")
    data = ones(20)
    S1 = LumiSpectrum(data, axes=(axis.get_axis_dictionary(),))
    with raises(AttributeError, match="Laser wavelength"):
        S1.to_raman_shift()
    with raises(AttributeError, match="Laser wavelength units"):
        S1.to_raman_shift(laser=0.244)
    S1.metadata.set_item("Acquisition_instrument.Laser.wavelength", 244)
    S2 = S1.to_raman_shift(inplace=False)
    S1.axes_manager[0].units = "µm"
    S1.axes_manager[0].axis = axis.axis / 1000
    with raises(AttributeError, match="Laser wavelength units"):
        S1.to_raman_shift(laser=244)
    S1.metadata.set_item("Acquisition_instrument.Laser.wavelength", 0.244)
    S1.to_raman_shift()
    assert S1.axes_manager[0].units == r"cm$^{-1}$"
    assert S2.axes_manager[0].name == "Raman Shift"
    assert S2.axes_manager[0].size == 20
    assert S1.axes_manager[0].axis[0] == S2.axes_manager[0].axis[0]
    assert_allclose(S1.data, S2.data, 5e-4)


def test_ramanshift_slicing():
    S = LumiSpectrum(arange(100), axes=[{"axis": arange(100) + 300}])
    S.to_raman_shift(inplace=True, laser=295)
    S.isig[3000.0:8000.0]


def test_solve_grating_equation():
    with warns(SyntaxWarning, match="(not in pixel units)"):
        axis = UniformDataAxis(size=20, offset=200, scale=10, units="nm")
        solve_grating_equation(axis, 3, -20, 300, 25, 600, 150)

    axis1 = UniformDataAxis(
        size=10,
        offset=200,
        scale=10,
    )
    axis2 = UniformDataAxis(size=10, offset=200, scale=10, units="px")

    nm_axis1 = solve_grating_equation(axis1, 3, -20, 300, 25, 600, 150)
    with warns(UserWarning, match="(range exceeds)"):
        nm_axis2 = solve_grating_equation(axis2, 1, 1, 1, 1, 1, 1)

    assert nm_axis1.name == "Wavelength"
    assert nm_axis1.units == "nm"
    assert nm_axis2.name == "Wavelength"
    assert nm_axis2.units == "nm"

    assert_allclose(nm_axis1.axis[0], 368.614, atol=0.1)
    assert_allclose(nm_axis1.axis[-1], 768.249, atol=0.1)
    assert_allclose(nm_axis2.axis[0], 411559.839, atol=0.1)
    assert_allclose(nm_axis2.axis[-1], 321785.967, atol=0.1)



import numpy as np
import pytest
from numpy.testing import assert_allclose
from lumispy.utils.signals import com
from hyperspy.axes import FunctionalDataAxis, DataAxis, UniformDataAxis


@pytest.mark.parametrize(
    "axis, output",
    [
        (
            FunctionalDataAxis(
                **{
                    "expression": "a * x + b",
                    "a": 1,
                    "b": 0,
                },
                size=2,
            ),
            0.5,
        ),
        (DataAxis(axis=[0.0, 1.0]), 0.5),
        (UniformDataAxis(size=2), 0.5),
    ],
)
def test_com_axes(axis, output):
    intensities = np.array([1, 1])

    centroid = com(intensities, axis)
    assert_allclose(centroid, output, atol=0.1)


def test_com_list():
    # Float without decimals as index for centroid
    wavelengths = [200, 300, 400, 500, 600, 700]
    intensities = np.array([1, 2, 3, 2, 1, 0])
    centroid = com(intensities, wavelengths)
    assert_allclose(centroid, 400.0, atol=0.1)

    # Float with decimals as index for centroid
    wavelengths = [
        200,
        300,
    ]
    intensities = np.array(
        [
            1,
            1,
        ]
    )
    centroid = com(intensities, wavelengths)
    assert_allclose(centroid, 250.0, atol=0.1)


def test_com_inputs():
    with pytest.raises(ValueError, match="The length of the spectrum array"):
        com(np.ones(2), np.ones(3))
    with pytest.raises(ValueError):
        com(np.ones(3), np.ones(2))
    with pytest.raises(
        ValueError, match="The parmeter `signal_axis` must be a HyperSpy Axis object."
    ):
        com(np.ones(3), "string")



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

SAVETXT_DOCSTRING = """
    Writes single spectra to a two-column data file with signal axis as
    X and data as Y.
    Writes linescan data to file with signal axis as first row and
    navigation axis as first column (flipped if `transpose=True`)."""

SAVETXT_PARAMETERS = """
    Parameters
    ----------
    filename : string
    fmt : str or sequence of strs, optional
        A single or sequence of format strings. Default is '%.5f'.
    delimiter : str, optional
        String or character separating columns. Default is ','
    axes : bool, optional
        If True (default), include axes in saved file. If False, save the data
        array only.
    transpose : bool, optional
        If True, transpose data array and exchange axes. Default is false.
        Ignored for single spectra.
    **kwargs 
        Takes any additional arguments of numpy.loadtxt, e.g. `newline`
        `header`, `footer`, `comments`, or `encoding`.

    See also
    --------
    numpy.savetxt"""

SAVETXT_EXAMPLE = """
    Examples
    --------
    >>> import lumispy as lum
    >>> import numpy as np
    ...
    >>> # Spectrum:
    >>> S = lum.signals.LumiSpectrum(np.arange(5))
    >>> lum.savetxt(S, 'spectrum.txt')
    0.00000	0.00000
    1.00000	1.00000
    2.00000	2.00000
    3.00000	3.00000
    4.00000	4.00000
    ...
    >>> # Linescan:
    >>> L = lum.signals.LumiSpectrum(np.arange(25).reshape((5,5)))
    >>> lum.savetxt(L, 'linescan.txt')
    0.00000	0.00000	1.00000	2.00000	3.00000	4.00000
    0.00000	0.00000	5.00000	10.00000	15.00000	20.00000
    1.00000	1.00000	6.00000	11.00000	16.00000	21.00000
    2.00000	2.00000	7.00000	12.00000	17.00000	22.00000
    3.00000	3.00000	8.00000	13.00000	18.00000	23.00000
    4.00000	4.00000	9.00000	14.00000	19.00000	24.00000
    """

TOARRAY_DOCSTRING = """
    Returns single spectra as two-column array.
    Returns linescan data as array with signal axis as first row and
    navigation axis as first column (flipped if `transpose=True`)."""

TOARRAY_PARAMETERS = """
    Parameters
    ----------
    axes : bool, optional
        If True (default), include axes in array. If False, return the data
        array only.
    transpose : bool, optional
        If True, transpose data array and exchange axes. Default is false.
        Ignored for single spectra.
    **kwargs 
        Takes any additional arguments of numpy.loadtxt, e.g. `newline`
        `header`, `footer`, `comments`, or `encoding`."""

TOARRAY_EXAMPLE = """
    Notes
    -----
    The output of this function can be used to convert a signal object to a
    pandas dataframe, e.g. using `df = pd.Dataframe(lum.to_array(S))`.
    
    Examples
    --------
    >>> import lumispy as lum
    >>> import numpy as np
    ...
    >>> # Spectrum:
    >>> S = lum.signals.LumiSpectrum(np.arange(5))
    >>> lum.to_array(S)
    array([[0., 0.],
          [1., 1.],
          [2., 2.],
          [3., 3.],
          [4., 4.]])
    ...
    >>> # Linescan:
    >>> L = lum.signals.LumiSpectrum(np.arange(25).reshape((5,5)))
    >>> lum.to_array(L)
    array([[ 0.,  0.,  1.,  2.,  3.,  4.],
          [ 0.,  0.,  1.,  2.,  3.,  4.],
          [ 1.,  5.,  6.,  7.,  8.,  9.],
          [ 2., 10., 11., 12., 13., 14.],
          [ 3., 15., 16., 17., 18., 19.],
          [ 4., 20., 21., 22., 23., 24.]])"""


def to_array(S, axes=True, transpose=False):
    """Returns signal object as numpy array (optionally including the axes).
    %s
    Returns image as array with the navigation axes as first column and first
    row.
    Returns 2D data (e.g. map of a fit parameter value) as array with the signal
    axes as first column and first row.
    %s
    %s
    """
    nav_axes = S.axes_manager.navigation_axes
    sig_axes = S.axes_manager.signal_axes
    dim = len(nav_axes) + len(sig_axes)
    # Convert single spectrum or data with only navigation axis
    if dim == 1:
        if axes:
            if len(sig_axes) == 1:
                x = sig_axes[0].axis
            else:
                x = nav_axes[0].axis
            output = np.array([x, S.data]).T
        else:
            output = S.data
    # Convert linescan or matrix
    elif dim == 2:
        if axes:
            if len(sig_axes) == 1:
                x = np.concatenate(([0], sig_axes[0].axis))
                y = nav_axes[0].axis
            elif len(nav_axes) == 0:
                x = np.concatenate(([0], sig_axes[0].axis))
                y = sig_axes[1].axis
            else:
                x = np.concatenate(([0], nav_axes[0].axis))
                y = nav_axes[1].axis
            if transpose:
                output = np.hstack(
                    (x.reshape(x.size, 1), np.vstack((y, np.transpose(S.data))))
                )
            else:
                output = np.vstack((x, np.hstack((y.reshape(y.size, 1), S.data))))
        else:
            if transpose:
                output = np.transpose(S.data)
            else:
                output = S.data
    else:
        raise NotImplementedError(
            "The to_array function currently handles a " "maximum of two axes."
        )
    return output


to_array.__doc__ %= (TOARRAY_DOCSTRING, TOARRAY_PARAMETERS, TOARRAY_EXAMPLE)


def savetxt(
    S, filename, fmt="%.5f", delimiter="\t", axes=True, transpose=False, **kwargs
):
    """Writes signal object to simple text file.
    %s
    Writes image to file with the navigation axes as first column and first
    row.
    Writes 2D data (e.g. map of a fit parameter value) to file with the signal
    axes as first column and first row.
    %s
    %s
    """
    dim = len(S.axes_manager.signal_axes) + len(S.axes_manager.navigation_axes)
    if dim <= 2:
        output = to_array(S, axes, transpose)
        np.savetxt(filename, output, fmt=fmt, delimiter=delimiter, **kwargs)
    else:
        raise NotImplementedError(
            "The savetxt function currently handles a " "maximum of two axes."
        )


savetxt.__doc__ %= (SAVETXT_DOCSTRING, SAVETXT_PARAMETERS, SAVETXT_EXAMPLE)



# Configuration file for the Sphinx documentation builder.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
import sys
from datetime import datetime
from importlib.metadata import version as get_version

sys.path.append("../")


# -- Project information
project = "LumiSpy"
copyright = f"2019-{datetime.today().year}, The LumiSpy developers"
author = "The LumiSpy developers"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_version("lumispy")
# The short X.Y version.
version = ".".join(release.split(".")[:2])

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.graphviz",
    "sphinx.ext.autosummary",
    "sphinxcontrib.towncrier",
    "sphinx_copybutton",
]

intersphinx_mapping = {
    "exspy": ("https://hyperspy.org/exspy/", None),
    "hyperspy": ("https://hyperspy.org/hyperspy-doc/current/", None),
    "kikuchipy": ("https://kikuchipy.org/en/latest/", None),
    "rsciio": ("https://hyperspy.org/rosettasciio/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}
intersphinx_disabled_domains = ["std"]

linkcheck_ignore = [
    "https://doi.org/10.1063/5.0080486",  # 403 Client Error: Forbidden for url
    "https://doi.org/10.1021/jz401508t",  # 403 Client Error: Forbidden for url
    "https://github.com/LumiSpy/lumispy/security/code-scanning",  # 404 Client Error: Not Found for url (even though page exists)
]

# imgmath: Sphinx allows use of LaTeX in the html documentation, but not directly. It is first rendered to an image.
# You can add here whatever preamble you are used to adding to your LaTeX document.
imgmath_latex_preamble = r"""
    \usepackage{xcolor}
    \definecolor{mathcolor}{rgb}{0.8,0.3,0.1}
    \everymath{\color{mathcolor}}
    %\everydisplay{\color{mathcolor}}
"""


templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "logo_only": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo_rec_april21.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/lumispy.ico"

# -- Options for EPUB output
epub_show_urls = "footnote"


def run_apidoc(_):
    # https://www.sphinx-doc.org/en/master/man/sphinx-apidoc.html
    # https://www.sphinx-doc.org/es/1.2/ext/autodoc.html
    import os

    os.environ["SPHINX_APIDOC_OPTIONS"] = (
        "members,private-members,no-undoc-members,show-inheritance,ignore-module-all"
    )

    from sphinx.ext.apidoc import main

    cur_dir = os.path.normpath(os.path.dirname(__file__))
    output_path = os.path.join(cur_dir, "api")
    modules = os.path.normpath(os.path.join(cur_dir, "../lumispy"))
    exclude_pattern = [
        "../lumispy/tests",
        "../lumispy/components",
        "../lumispy/release_info.py",
    ]
    main(["-e", "-f", "-P", "-o", output_path, modules, *exclude_pattern])


def setup(app):
    app.connect("builder-inited", run_apidoc)
    app.add_css_file("css/dark.css")
    app.add_css_file("css/light.css")


# CopyButton configuration
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"


# -- Options for towncrier_draft extension -----------------------------------

# Options: draft/sphinx-version/sphinx-release
towncrier_draft_autoversion_mode = "draft"
towncrier_draft_include_empty = False
towncrier_draft_working_directory = ".."



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

from lumispy.signals import LumiSpectrum, LumiTransient, LumiTransientSpectrum


class TestLumiTransientSpectrum0D:
    def setup_method(self, method):
        ax0 = {"name": "Time", "units": "ps", "size": 10}
        ax1 = {"name": "Wavelength", "units": "nm", "size": 10}
        self.s = LumiTransientSpectrum(np.ones(100).reshape(10, 10), axes=[ax0, ax1])

    def test_slice_wavelength(self):
        s2 = self.s.isig[:, 5]
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_slice_energy(self):
        s2 = self.s.to_eV(inplace=False)
        s3 = s2.isig[:, 5]
        assert s3.axes_manager[-1].units == "eV"
        assert type(s3) == LumiSpectrum

    def test_slice_time(self):
        s2 = self.s.isig[5, :]
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_sum_wl(self):
        s2 = self.s.sum(axis="Wavelength")
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_sum_E(self):
        s2 = self.s.to_eV(inplace=False)
        s3 = s2.sum(axis="Energy")
        assert s3.axes_manager[-1].units == "ps"
        assert type(s3) == LumiTransient

    def test_sum_t(self):
        s2 = self.s.sum(axis="Time")
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_max_wl(self):
        s2 = self.s.max(axis="Wavelength")
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_max_t(self):
        s2 = self.s.max(axis="Time")
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_spec2nav(self):
        s2 = self.s.spec2nav()
        assert s2.axes_manager[0].units == "nm"
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_time2nav(self):
        s2 = self.s.time2nav()
        assert s2.axes_manager[0].units == "ps"
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum


class TestLumiTransientSpectrum2D:
    def setup_method(self, method):
        nav = {"size": 10, "navigate": True}
        ax0 = {"name": "Time", "units": "ps", "size": 10}
        ax1 = {"name": "Wavelength", "units": "nm", "size": 10}
        self.s = LumiTransientSpectrum(
            np.ones(10000).reshape(10, 10, 10, 10), axes=[nav, nav, ax0, ax1]
        )

    def test_slice_wavelength(self):
        s2 = self.s.isig[:, 5]
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_slice_energy(self):
        s2 = self.s.to_eV(inplace=False)
        s3 = s2.isig[:, 5]
        assert s3.axes_manager[-1].units == "eV"
        assert type(s3) == LumiSpectrum

    def test_slice_time(self):
        s2 = self.s.isig[5, :]
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_sum_wl(self):
        s2 = self.s.sum(axis="Wavelength")
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_sum_E(self):
        s2 = self.s.to_eV(inplace=False)
        s3 = s2.sum(axis="Energy")
        assert s3.axes_manager[-1].units == "ps"
        assert type(s3) == LumiTransient

    def test_sum_t(self):
        s2 = self.s.sum(axis="Time")
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_max_wl(self):
        s2 = self.s.max(axis="Wavelength")
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_max_t(self):
        s2 = self.s.max(axis="Time")
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum

    def test_spec2nav(self):
        s2 = self.s.spec2nav()
        assert s2.axes_manager[0].units == "nm"
        assert s2.axes_manager[-1].units == "ps"
        assert type(s2) == LumiTransient

    def test_time2nav(self):
        s2 = self.s.time2nav()
        assert s2.axes_manager[0].units == "ps"
        assert s2.axes_manager[-1].units == "nm"
        assert type(s2) == LumiSpectrum



# -*- coding: utf-8 -*-
# Copyright 2019-2023 The LumiSpy developers
#
# This file is part of LumiSpy.
#
# LumiSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# LumiSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LumiSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

from lumispy.signals import LumiSpectrum, LumiTransientSpectrum


class TestCommonLumi:
    def test_crop_edges(self):
        s1 = LumiSpectrum(np.ones((10, 10, 10)))
        s2 = LumiTransientSpectrum(np.ones((10, 10, 10, 10)))
        s3 = LumiSpectrum(np.ones((3, 3, 10)))
        s1 = s1.crop_edges(crop_px=2)
        s2 = s2.crop_edges(crop_px=2)
        assert s1.axes_manager.navigation_shape[0] == 6
        assert s1.axes_manager.navigation_shape[1] == 6
        assert s2.axes_manager.navigation_shape[0] == 6
        assert s2.axes_manager.navigation_shape[1] == 6
        with pytest.raises(ValueError):
            s3.crop_edges(crop_px=2)

    def test_remove_negative(self):
        s1 = LumiSpectrum(np.random.random((10, 10, 10))) - 0.3
        s2 = LumiTransientSpectrum(np.random.random((10, 10, 10, 10))) - 0.3
        s3 = LumiTransientSpectrum(np.random.random((10, 10, 10, 10))) - 0.3
        s1a = s1.remove_negative(inplace=False)
        s2a = s2.remove_negative(inplace=False)
        s3a = s3.remove_negative(basevalue=0.1)
        assert s3a.metadata.Signal.negative_removed == True
        assert np.all(s1a.data[s1 <= 0] == 1)
        assert np.all(s2a.data[s2 <= 0] == 1)
        assert np.all(s3a.data[s3 <= 0] == 0.1)
        s1.remove_negative(inplace=True)
        s2.remove_negative(inplace=True)
        s3.remove_negative(basevalue=0.1, inplace=True)
        assert s1 == s1a
        assert s2 == s2a
        assert s3 == s3a

    def test_scale_by_exposure(self):
        s1 = LumiSpectrum(np.ones((10, 10, 10)))
        s2 = LumiTransientSpectrum(np.ones((10, 10, 10, 10)))
        s2.axes_manager.signal_axes[-1].units = "ns"
        s2.axes_manager.signal_axes[0].units = "nm"
        s4 = LumiSpectrum(np.ones((10)))
        s2.metadata.set_item("Acquisition_instrument.Detector.integration_time", 2)
        s2.metadata.set_item("Signal.quantity", "Intensity (Counts)")
        s4.metadata.set_item("Signal.quantity", "Intensity (counts)")
        s1a = s1.scale_by_exposure(integration_time=4)
        s2a = s2.scale_by_exposure()
        s4a = s4.scale_by_exposure(integration_time=0.1)
        assert np.all(s1a.data == 0.25)
        assert np.all(s2a.data == 0.5)
        assert np.all(s4a.data == 10)
        assert s2a.metadata.Signal.quantity == "Intensity (Counts/s)"
        assert s4a.metadata.Signal.quantity == "Intensity (counts/s)"
        assert s4a.metadata.Signal.scaled == True
        s1.scale_by_exposure(integration_time=4, inplace=True)
        s2.scale_by_exposure(inplace=True)
        s4.scale_by_exposure(integration_time=0.1, inplace=True)
        assert s1 == s1a
        assert s2 == s2a
        assert s4 == s4a
        assert s2.metadata.Signal.quantity == "Intensity (Counts/s)"
        assert s4.metadata.Signal.quantity == "Intensity (counts/s)"
        # Test for errors
        s4 = LumiSpectrum(np.ones((10)))
        s4.normalize(inplace=True)
        with pytest.raises(AttributeError, match="Data was normalized and"):
            s4.scale_by_exposure(inplace=True, integration_time=0.5)
        s5 = LumiSpectrum(np.ones((10)))
        with pytest.raises(AttributeError, match="not included in the"):
            s5.scale_by_exposure(inplace=True)
        s5.scale_by_exposure(inplace=True, integration_time=0.5)
        with pytest.raises(AttributeError, match="Data was already scaled."):
            s5.scale_by_exposure(inplace=True, integration_time=0.5)
        # Deprecation test for exposure argument
        s6 = LumiSpectrum(np.ones((10)))
        with pytest.warns(DeprecationWarning, match="removed in LumiSpy 1.0"):
            s6.scale_by_exposure(inplace=True, exposure=0.5)
            assert np.all(s6.data == 2)

    def test_normalize(self):
        s1 = LumiSpectrum(np.random.random((10, 10, 10))) * 2
        s2 = LumiTransientSpectrum(np.random.random((10, 10, 10, 10))) * 2
        s2.axes_manager.signal_axes[-1].units = "ps"
        s2.axes_manager.signal_axes[0].units = "nm"
        s3 = LumiSpectrum(np.random.random((10, 10))) * 2
        s4 = LumiSpectrum(np.random.random((10))) * 2
        s4.metadata.set_item("Signal.quantity", "Intensity (counts)")
        s1a = s1.normalize()
        s2a = s2.normalize()
        s3a = s3.normalize()
        s4a = s4.normalize()
        assert s1a.max(axis=[0, 1, 2]).data[0] == 1
        assert s2a.max(axis=[0, 1, 2, 3]).data[0] == 1
        assert s3a.max(axis=[0, 1]).data[0] == 1
        assert s4a.max(axis=[0]).data[0] == 1
        assert s4a.metadata.Signal.quantity == "Normalized intensity"
        assert s4a.metadata.Signal.normalized == True
        s1a = s1.normalize(element_wise=True)
        s2a = s2.normalize(element_wise=True)
        s3a = s3.normalize(element_wise=True)
        s4a = s4.normalize(element_wise=True)
        assert np.all(s1a.max(axis=[2]).data[0] == 1)
        assert np.all(s2a.max(axis=[3]).data[0] == 1)
        assert np.all(s3a.max(axis=[1]).data[0] == 1)
        assert s4a.max(axis=[0]).data[0] == 1
        s1a = s1.normalize(pos=3)
        s2a = s2.normalize(pos=3, element_wise=True)
        s3a = s3.normalize(pos=3, element_wise=True)
        s4a = s4.normalize(pos=3)
        assert s1a.isig[3].max().data[0] == 1
        assert np.all(s2a.isig[3] == 1)
        assert np.all(s3a.isig[3] == 1)
        assert s4a.isig[3].data == 1
        s1.normalize(pos=3, inplace=True)
        s2.normalize(pos=3, element_wise=True, inplace=True)
        s3.normalize(pos=3, element_wise=True, inplace=True)
        s4.normalize(pos=3, inplace=True)
        assert s1a == s1
        assert s2a == s2
        assert s3a == s3
        assert s4a == s4
        assert s4.metadata.Signal.quantity == "Normalized intensity"
        with pytest.warns(UserWarning, match="Data was"):
            s1.normalize(inplace=True)



import numpy as np
from hyperspy.axes import FunctionalDataAxis
from scipy.ndimage import center_of_mass
from scipy.interpolate import interp1d


def com(spectrum_intensities, signal_axis, **kwargs):
    """Finds the centroid (center of mass) of a peak in the spectrum based
    from the intensity at each pixel value and its respective signal axis.

    Parameters
    ----------
    spectrum_intensities : array
        An array with the intensities of the spectrum.
    signal_axis: hyperspy.axes.BaseDataAxis subclass
        A HyperSpy signal axis class containing an array with the wavelength/
        energy for each intensity/signal value.
    kwargs : dictionary
        For the scipy.interpolate.interp1d function.

    Returns
    -------
    center_of_mass : float
        The centroid of the spectrum.

    Examples
    --------
    # Assume we have a spectrum with wavelengths and intensities
    >>> wavelengths = [200, 300, 400, 500, 600, 700]
    >>> intensities = [1, 2, 3, 2, 1, 0]
    >>> from hyperspy.axes import DataAxis
    >>> signal_axis = DataAxis(axis=wavelengths)

    >>> center_of_mass = com(intensities, signal_axis)
    >>> print(center_of_mass)  # Outputs: [400.0]
    """

    def _interpolate_signal(axis_array, index, **kwargs):
        """
        Wrapper for `hs.axes.index2value` that linearly interpolates between
        values should the index passed not be a integer. Using the kwargs, the
        interpolation method can be changed.
        """
        rem = index % 1
        index = int(index // 1)
        if rem == 0:
            return axis_array[int(index)]
        else:
            y = [axis_array[index], axis_array[index + 1]]
            x = [0, 1]
            fx = interp1d(x, y, **kwargs)
            return float(fx(rem))

    # Find center of mass wrt array index
    index_com = float(center_of_mass(spectrum_intensities)[0])

    # Check for the type of hyperspy.axis
    if type(signal_axis) == FunctionalDataAxis:
        # Calculate value y from x[index_com]
        x = _interpolate_signal(signal_axis.x.axis, index_com)
        kwargs = {}
        for kwarg in signal_axis.parameters_list:
            kwargs[kwarg] = getattr(signal_axis, kwarg)
        com_val = signal_axis._function(x, **kwargs)

    elif hasattr(signal_axis, "axis"):
        # Calculate value interpolating between index_com 0 and 1
        com_val = _interpolate_signal(signal_axis.axis, index_com)
    elif type(signal_axis) in (list, np.ndarray, tuple):
        # Check for dimensionality
        if len(spectrum_intensities) != len(signal_axis):
            raise ValueError(
                f"The length of the spectrum array {len(spectrum_intensities)} must match "
                "the length of the wavelength array {len(signal_axis)}."
            )
        # Calculate value interpolating between index_com 0 and 1
        com_val = _interpolate_signal(np.array(signal_axis), index_com)
    else:
        raise ValueError("The parmeter `signal_axis` must be a HyperSpy Axis object.")

    return com_val
